# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend.config import keras_3
from keras_cv.src.layers.spatial_pyramid import SpatialPyramidPooling
from keras_cv.src.models.backbones.backbone_presets import backbone_presets
from keras_cv.src.models.backbones.backbone_presets import (
    backbone_presets_with_weights,
)
from keras_cv.src.models.segmentation.deeplab_v3_plus.deeplab_v3_plus_presets import (  # noqa: E501
    deeplab_v3_plus_presets,
)
from keras_cv.src.models.task import Task
from keras_cv.src.utils.python_utils import classproperty
from keras_cv.src.utils.train import get_feature_extractor


@keras_cv_export(
    [
        "keras_cv.models.DeepLabV3Plus",
        "keras_cv.models.segmentation.DeepLabV3Plus",
    ]
)
class DeepLabV3Plus(Task):
    """A Keras model implementing the DeepLabV3+ architecture for semantic
    segmentation.

    References:
        - [Encoder-Decoder with Atrous Separable Convolution for Semantic Image Segmentation](https://arxiv.org/abs/1802.02611)  # noqa: E501
        (ECCV 2018)
        - [Rethinking Atrous Convolution for Semantic Image Segmentation](https://arxiv.org/abs/1706.05587)  # noqa: E501
        (CVPR 2017)

    Args:
        backbone: `keras.Model`. The backbone network for the model that is
            used as a feature extractor for the DeepLabV3+ Encoder. Should
            either be a `keras_cv.models.backbones.backbone.Backbone` or a
            `keras.Model` that implements the `pyramid_level_inputs`
            property with keys "P2" and "P5" and layer names as values. A
            somewhat sensible backbone to use in many cases is the
            `keras_cv.models.ResNet50V2Backbone.from_preset("resnet50_v2_imagenet")`.
        num_classes: int, the number of classes for the detection model. Note
            that the `num_classes` contains the background class, and the
            classes from the data should be represented by integers with range
            [0, `num_classes`).
        projection_filters: int, number of filters in the convolution layer
            projecting low-level features from the `backbone`. The default
            value is set to `48`, as per the
            [TensorFlow implementation of DeepLab](https://github.com/tensorflow/models/blob/master/research/deeplab/model.py#L676).  # noqa: E501
        spatial_pyramid_pooling: (Optional) a `keras.layers.Layer`. Also known
            as Atrous Spatial Pyramid Pooling (ASPP). Performs spatial pooling
            on different spatial levels in the pyramid, with dilation. If
            provided, the feature map from the backbone is passed to it inside
            the DeepLabV3 Encoder, otherwise
            `keras_cv.layers.spatial_pyramid.SpatialPyramidPooling` is used.
        segmentation_head: (Optional) a `keras.layers.Layer`. If provided, the
            outputs of the DeepLabV3 encoder is passed to this layer and it
            should predict the segmentation mask based on feature from backbone
            and feature from decoder, otherwise a default DeepLabV3
            convolutional head is used.

    Example:
    ```python
    import keras_cv

    images = np.ones(shape=(1, 96, 96, 3))
    labels = np.zeros(shape=(1, 96, 96, 1))
    backbone = keras_cv.models.ResNet50V2Backbone(input_shape=[96, 96, 3])
    model = keras_cv.models.segmentation.DeepLabV3Plus(
        num_classes=1, backbone=backbone,
    )

    # Evaluate model
    model(images)

    # Train model
    model.compile(
        optimizer="adam",
        loss=keras.losses.BinaryCrossentropy(from_logits=False),
        metrics=["accuracy"],
    )
    model.fit(images, labels, epochs=3)
    ```
    """

    def __init__(
        self,
        backbone,
        num_classes,
        projection_filters=48,
        spatial_pyramid_pooling=None,
        segmentation_head=None,
        **kwargs,
    ):
        if not isinstance(backbone, keras.layers.Layer) or not isinstance(
            backbone, keras.Model
        ):
            raise ValueError(
                "Argument `backbone` must be a `keras.layers.Layer` instance "
                f" or `keras.Model`. Received instead "
                f"backbone={backbone} (of type {type(backbone)})."
            )

        inputs = backbone.input

        extractor_levels = ["P2", "P5"]
        extractor_layer_names = [
            backbone.pyramid_level_inputs[i] for i in extractor_levels
        ]
        feature_extractor = get_feature_extractor(
            backbone, extractor_layer_names, extractor_levels
        )
        backbone_features = feature_extractor(inputs)

        if spatial_pyramid_pooling is None:
            spatial_pyramid_pooling = SpatialPyramidPooling(
                dilation_rates=[6, 12, 18]
            )
        spp_outputs = spatial_pyramid_pooling(backbone_features["P5"])

        low_level_feature_projector = keras.Sequential(
            [
                keras.layers.Conv2D(
                    name="low_level_feature_conv",
                    filters=projection_filters,
                    kernel_size=1,
                    padding="same",
                    use_bias=False,
                ),
                keras.layers.BatchNormalization(name="low_level_feature_norm"),
                keras.layers.ReLU(name="low_level_feature_relu"),
            ]
        )

        low_level_projected_features = low_level_feature_projector(
            backbone_features["P2"]
        )

        encoder_outputs = keras.layers.UpSampling2D(
            size=(8, 8),
            interpolation="bilinear",
            name="encoder_output_upsampling",
        )(spp_outputs)

        combined_encoder_outputs = keras.layers.Concatenate(axis=-1)(
            [encoder_outputs, low_level_projected_features]
        )

        if segmentation_head is None:
            segmentation_head = keras.Sequential(
                [
                    keras.layers.Conv2D(
                        name="segmentation_head_conv",
                        filters=256,
                        kernel_size=1,
                        padding="same",
                        use_bias=False,
                    ),
                    keras.layers.BatchNormalization(
                        name="segmentation_head_norm"
                    ),
                    keras.layers.ReLU(name="segmentation_head_relu"),
                    keras.layers.UpSampling2D(
                        size=(4, 4), interpolation="bilinear"
                    ),
                    # Classification layer
                    keras.layers.Conv2D(
                        name="segmentation_output",
                        filters=num_classes,
                        kernel_size=1,
                        use_bias=False,
                        padding="same",
                        activation="softmax",
                        # Force the dtype of the classification layer to float32
                        # to avoid the NAN loss issue when used with mixed
                        # precision API.
                        dtype="float32",
                    ),
                ]
            )

        outputs = segmentation_head(combined_encoder_outputs)

        super().__init__(inputs=inputs, outputs=outputs, **kwargs)

        self.num_classes = num_classes
        self.backbone = backbone
        self.spatial_pyramid_pooling = spatial_pyramid_pooling
        self.projection_filters = projection_filters
        self.segmentation_head = segmentation_head

    def get_config(self):
        return {
            "num_classes": self.num_classes,
            "backbone": keras.saving.serialize_keras_object(self.backbone),
            "spatial_pyramid_pooling": keras.saving.serialize_keras_object(
                self.spatial_pyramid_pooling
            ),
            "projection_filters": self.projection_filters,
            "segmentation_head": keras.saving.serialize_keras_object(
                self.segmentation_head
            ),
        }

    @classmethod
    def from_config(cls, config):
        if "backbone" in config and isinstance(config["backbone"], dict):
            config["backbone"] = keras.layers.deserialize(config["backbone"])
        if "spatial_pyramid_pooling" in config and isinstance(
            config["spatial_pyramid_pooling"], dict
        ):
            config["spatial_pyramid_pooling"] = keras.layers.deserialize(
                config["spatial_pyramid_pooling"]
            )
        if "segmentation_head" in config and isinstance(
            config["segmentation_head"], dict
        ):
            config["segmentation_head"] = keras.layers.deserialize(
                config["segmentation_head"]
            )
        return super().from_config(config)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        if keras_3():
            return copy.deepcopy(
                {**backbone_presets, **deeplab_v3_plus_presets}
            )
        else:
            # TODO: #2246 Deeplab V3 presets don't work in Keras 2
            return copy.deepcopy({**backbone_presets})

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy(
            {**backbone_presets_with_weights, **deeplab_v3_plus_presets}
        )

    @classproperty
    def backbone_presets(cls):
        """Dictionary of preset names and configurations of compatible
        backbones."""
        return copy.deepcopy(backbone_presets)



# Copyright 2024 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.models.backbones.video_swin.video_swin_backbone import (
    VideoSwinBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.utils.python_utils import classproperty

ALIAS_DOCSTRING = """VideoSwin{size}Backbone model.

    Reference:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)

    For transfer learning use cases, make sure to read the
    [guide to transfer learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Examples:
    ```python
    input_data = np.ones(shape=(1, 32, 224, 224, 3))

    # Randomly initialized backbone
    model = VideoSwin{size}Backbone()
    output = model(input_data)
    ```
"""  # noqa: E501


@keras_cv_export("keras_cv.models.VideoSwinTBackbone")
class VideoSwinTBackbone(VideoSwinBackbone):
    def __new__(
        cls,
        embed_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[8, 7, 7],
        include_rescaling=True,
        **kwargs,
    ):
        kwargs.update(
            {
                "embed_dim": embed_dim,
                "depths": depths,
                "num_heads": num_heads,
                "window_size": window_size,
                "include_rescaling": include_rescaling,
            }
        )
        return VideoSwinBackbone.from_preset("videoswin_tiny", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "videoswin_tiny_kinetics400": copy.deepcopy(
                backbone_presets["videoswin_tiny_kinetics400"]
            ),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.VideoSwinSBackbone")
class VideoSwinSBackbone(VideoSwinBackbone):
    def __new__(
        cls,
        embed_dim=96,
        depths=[2, 2, 18, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[8, 7, 7],
        include_rescaling=True,
        **kwargs,
    ):
        kwargs.update(
            {
                "embed_dim": embed_dim,
                "depths": depths,
                "num_heads": num_heads,
                "window_size": window_size,
                "include_rescaling": include_rescaling,
            }
        )
        return VideoSwinBackbone.from_preset("videoswin_small", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "videoswin_small_kinetics400": copy.deepcopy(
                backbone_presets["videoswin_small_kinetics400"]
            ),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.VideoSwinBBackbone")
class VideoSwinBBackbone(VideoSwinBackbone):
    def __new__(
        cls,
        embed_dim=128,
        depths=[2, 2, 18, 2],
        num_heads=[4, 8, 16, 32],
        window_size=[8, 7, 7],
        include_rescaling=True,
        **kwargs,
    ):
        kwargs.update(
            {
                "embed_dim": embed_dim,
                "depths": depths,
                "num_heads": num_heads,
                "window_size": window_size,
                "include_rescaling": include_rescaling,
            }
        )
        return VideoSwinBackbone.from_preset("videoswin_base", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "videoswin_base_kinetics400": copy.deepcopy(
                backbone_presets["videoswin_base_kinetics400"]
            ),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


setattr(VideoSwinTBackbone, "__doc__", ALIAS_DOCSTRING.format(size="T"))
setattr(VideoSwinSBackbone, "__doc__", ALIAS_DOCSTRING.format(size="S"))
setattr(VideoSwinBBackbone, "__doc__", ALIAS_DOCSTRING.format(size="B"))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export("keras_cv.losses.GIoULoss")
class GIoULoss(keras.losses.Loss):
    """Implements the Generalized IoU Loss

    GIoU loss is a modified IoU loss commonly used for object detection. This
    loss aims to directly optimize the IoU score between true boxes and
    predicted boxes. GIoU loss adds a penalty term to the IoU loss that takes in
    account the area of the smallest box enclosing both the boxes being
    considered for the iou. The length of the last dimension should be 4 to
    represent the bounding boxes.

    Args:
        bounding_box_format: a case-insensitive string (for example, "xyxy").
            Each bounding box is defined by these 4 values.For detailed
            information on the supported formats, see the [KerasCV bounding box
            documentation](https://keras.io/api/keras_cv/bounding_box/formats/).
        axis: the axis along which to mean the ious, defaults to -1.

    References:
        - [GIoU paper](https://arxiv.org/pdf/1902.09630)
        - [TFAddons Implementation](https://www.tensorflow.org/addons/api_docs/python/tfa/losses/GIoULoss)

    Example:
    ```python
    y_true = np.random.uniform(size=(5, 10, 5), low=0, high=10)
    y_pred = np.random.uniform(size=(5, 10, 4), low=0, high=10)
    loss = GIoULoss(bounding_box_format = "xyWH")
    loss(y_true, y_pred).numpy()
    ```

    Usage with the `compile()` API:
    ```python
    model.compile(optimizer='adam', loss=keras_cv.losses.GIoULoss())
    ```
    """  # noqa: E501

    def __init__(self, bounding_box_format, axis=-1, **kwargs):
        super().__init__(**kwargs)
        self.bounding_box_format = bounding_box_format
        self.axis = axis

    def _compute_enclosure(self, boxes1, boxes2):
        y_min1, x_min1, y_max1, x_max1 = ops.split(boxes1[..., :4], 4, axis=-1)
        y_min2, x_min2, y_max2, x_max2 = ops.split(boxes2[..., :4], 4, axis=-1)
        boxes2_rank = len(boxes2.shape)
        perm = [1, 0] if boxes2_rank == 2 else [0, 2, 1]
        # [N, M] or [batch_size, N, M]
        zeros_t = ops.cast(0, boxes1.dtype)

        enclose_ymin = ops.minimum(y_min1, ops.transpose(y_min2, perm))
        enclose_xmin = ops.minimum(x_min1, ops.transpose(x_min2, perm))
        enclose_ymax = ops.maximum(y_max1, ops.transpose(y_max2, perm))
        enclose_xmax = ops.maximum(x_max1, ops.transpose(x_max2, perm))
        enclose_width = ops.maximum(zeros_t, enclose_xmax - enclose_xmin)
        enclose_height = ops.maximum(zeros_t, enclose_ymax - enclose_ymin)
        enclose_area = enclose_width * enclose_height

        return enclose_area

    def _compute_giou(self, boxes1, boxes2):
        boxes1_rank = len(boxes1.shape)
        boxes2_rank = len(boxes2.shape)

        if boxes1_rank not in [2, 3]:
            raise ValueError(
                "compute_iou() expects boxes1 to be batched, or to be "
                f"unbatched. Received len(boxes1.shape)={boxes1_rank}, "
                f"len(boxes2.shape)={boxes2_rank}. Expected either "
                "len(boxes1.shape)=2 AND or len(boxes1.shape)=3."
            )
        if boxes2_rank not in [2, 3]:
            raise ValueError(
                "compute_iou() expects boxes2 to be batched, or to be "
                f"unbatched. Received len(boxes1.shape)={boxes1_rank}, "
                f"len(boxes2.shape)={boxes2_rank}. Expected either "
                "len(boxes2.shape)=2 AND or len(boxes2.shape)=3."
            )

        target_format = "yxyx"
        if bounding_box.is_relative(self.bounding_box_format):
            target_format = bounding_box.as_relative(target_format)

        boxes1 = bounding_box.convert_format(
            boxes1, source=self.bounding_box_format, target=target_format
        )

        boxes2 = bounding_box.convert_format(
            boxes2, source=self.bounding_box_format, target=target_format
        )

        intersect_area = bounding_box.iou._compute_intersection(boxes1, boxes2)
        boxes1_area = bounding_box.iou._compute_area(boxes1)
        boxes2_area = bounding_box.iou._compute_area(boxes2)
        boxes2_area_rank = len(boxes2_area.shape)
        boxes2_axis = 1 if (boxes2_area_rank == 2) else 0
        boxes1_area = ops.expand_dims(boxes1_area, axis=-1)
        boxes2_area = ops.expand_dims(boxes2_area, axis=boxes2_axis)
        union_area = boxes1_area + boxes2_area - intersect_area
        iou = ops.divide(intersect_area, union_area + keras.backend.epsilon())

        # giou calculation
        enclose_area = self._compute_enclosure(boxes1, boxes2)

        return iou - ops.divide(
            (enclose_area - union_area), enclose_area + keras.backend.epsilon()
        )

    def call(self, y_true, y_pred, sample_weight=None):
        if sample_weight is not None:
            raise ValueError(
                "GIoULoss does not support sample_weight. Please ensure "
                f"sample_weight=None. Got sample_weight={sample_weight}"
            )

        y_pred = ops.convert_to_tensor(y_pred)
        y_true = ops.cast(y_true, y_pred.dtype)

        if y_pred.shape[-1] != 4:
            raise ValueError(
                "GIoULoss expects y_pred.shape[-1] to be 4 to represent the "
                f"bounding boxes. Received y_pred.shape[-1]={y_pred.shape[-1]}."
            )

        if y_true.shape[-1] != 4:
            raise ValueError(
                "GIoULoss expects y_true.shape[-1] to be 4 to represent the "
                f"bounding boxes. Received y_true.shape[-1]={y_true.shape[-1]}."
            )

        if y_true.shape[-2] != y_pred.shape[-2]:
            raise ValueError(
                "GIoULoss expects number of boxes in y_pred to be equal to the "
                "number of boxes in y_true. Received number of boxes in "
                f"y_true={y_true.shape[-2]} and number of boxes in "
                f"y_pred={y_pred.shape[-2]}."
            )

        giou = self._compute_giou(y_true, y_pred)
        giou = ops.diagonal(
            giou,
        )
        if self.axis == "no_reduction":
            warnings.warn(
                "`axis='no_reduction'` is a temporary API, and the API "
                "contract will be replaced in the future with a more generic "
                "solution covering all losses."
            )
        else:
            giou = ops.mean(giou, axis=self.axis)

        return 1 - giou

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "bounding_box_format": self.bounding_box_format,
                "axis": self.axis,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for working with bounding boxes."""

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import ops
from keras_cv.src.bounding_box.formats import XYWH


@keras_cv_export("keras_cv.bounding_box.is_relative")
def is_relative(bounding_box_format):
    """A util to check if a bounding box format uses relative coordinates"""
    if (
        bounding_box_format.lower()
        not in bounding_box.converters.TO_XYXY_CONVERTERS
    ):
        raise ValueError(
            "`is_relative()` received an unsupported format for the argument "
            f"`bounding_box_format`. `bounding_box_format` should be one of "
            f"{bounding_box.converters.TO_XYXY_CONVERTERS.keys()}. "
            f"Got bounding_box_format={bounding_box_format}"
        )

    return bounding_box_format.startswith("rel")


@keras_cv_export("keras_cv.bounding_box.as_relative")
def as_relative(bounding_box_format):
    """A util to get the relative equivalent of a provided bounding box format.

    If the specified format is already a relative format,
    it will be returned unchanged.
    """

    if not is_relative(bounding_box_format):
        return "rel_" + bounding_box_format

    return bounding_box_format


def _relative_area(boxes, bounding_box_format):
    boxes = bounding_box.convert_format(
        boxes,
        source=bounding_box_format,
        target="rel_xywh",
    )
    widths = boxes[..., XYWH.WIDTH]
    heights = boxes[..., XYWH.HEIGHT]
    # handle corner case where shear performs a full inversion.
    return ops.where(
        ops.logical_and(widths > 0, heights > 0), widths * heights, 0.0
    )


@keras_cv_export("keras_cv.bounding_box.clip_to_image")
def clip_to_image(
    bounding_boxes, bounding_box_format, images=None, image_shape=None
):
    """clips bounding boxes to image boundaries.

    `clip_to_image()` clips bounding boxes that have coordinates out of bounds
    of an image down to the boundaries of the image. This is done by converting
    the bounding box to relative formats, then clipping them to the `[0, 1]`
    range. Additionally, bounding boxes that end up with a zero area have their
    class ID set to -1, indicating that there is no object present in them.

    Args:
        bounding_boxes: bounding box tensor to clip.
        bounding_box_format: the KerasCV bounding box format the bounding boxes
            are in.
        images: list of images to clip the bounding boxes to.
        image_shape: the shape of the images to clip the bounding boxes to.
    """
    boxes, classes = bounding_boxes["boxes"], bounding_boxes["classes"]

    boxes = bounding_box.convert_format(
        boxes,
        source=bounding_box_format,
        target="rel_xyxy",
        images=images,
        image_shape=image_shape,
    )
    boxes, classes, images, squeeze = _format_inputs(boxes, classes, images)
    x1, y1, x2, y2 = ops.split(boxes, 4, axis=-1)
    clipped_bounding_boxes = ops.concatenate(
        [
            ops.clip(x1, 0, 1),
            ops.clip(y1, 0, 1),
            ops.clip(x2, 0, 1),
            ops.clip(y2, 0, 1),
        ],
        axis=-1,
    )
    areas = _relative_area(
        clipped_bounding_boxes, bounding_box_format="rel_xyxy"
    )
    clipped_bounding_boxes = bounding_box.convert_format(
        clipped_bounding_boxes,
        source="rel_xyxy",
        target=bounding_box_format,
        images=images,
        image_shape=image_shape,
    )
    clipped_bounding_boxes = ops.where(
        ops.expand_dims(areas > 0.0, axis=-1), clipped_bounding_boxes, -1.0
    )
    classes = ops.where(areas > 0.0, classes, -1)
    nan_indices = ops.any(ops.isnan(clipped_bounding_boxes), axis=-1)
    classes = ops.where(nan_indices, -1, classes)

    # TODO update dict and return
    clipped_bounding_boxes, classes = _format_outputs(
        clipped_bounding_boxes, classes, squeeze
    )

    result = bounding_boxes.copy()
    result["boxes"] = clipped_bounding_boxes
    result["classes"] = classes
    return result


# TODO (tanzhenyu): merge with clip_to_image
def _clip_boxes(boxes, box_format, image_shape):
    """Clip boxes to the boundaries of the image shape"""
    if boxes.shape[-1] != 4:
        raise ValueError(
            "boxes.shape[-1] is {:d}, but must be 4.".format(boxes.shape[-1])
        )

    if isinstance(image_shape, list) or isinstance(image_shape, tuple):
        height, width, _ = image_shape
        max_length = [height, width, height, width]
    else:
        image_shape = ops.cast(image_shape, dtype=boxes.dtype)
        height = image_shape[0]
        width = image_shape[1]
        max_length = ops.stack([height, width, height, width], axis=-1)

    clipped_boxes = ops.maximum(ops.minimum(boxes, max_length), 0.0)
    return clipped_boxes


def _format_inputs(boxes, classes, images):
    boxes_rank = len(boxes.shape)
    if boxes_rank > 3:
        raise ValueError(
            "Expected len(boxes.shape)=2, or len(boxes.shape)=3, got "
            f"len(boxes.shape)={boxes_rank}"
        )
    boxes_includes_batch = boxes_rank == 3
    # Determine if images needs an expand_dims() call
    if images is not None:
        images_rank = len(images.shape)
        if images_rank > 4:
            raise ValueError(
                "Expected len(images.shape)=2, or len(images.shape)=3, got "
                f"len(images.shape)={images_rank}"
            )
        images_include_batch = images_rank == 4
        if boxes_includes_batch != images_include_batch:
            raise ValueError(
                "clip_to_image() expects both boxes and images to be batched, "
                "or both boxes and images to be unbatched. Received "
                f"len(boxes.shape)={boxes_rank}, "
                f"len(images.shape)={images_rank}. Expected either "
                "len(boxes.shape)=2 AND len(images.shape)=3, or "
                "len(boxes.shape)=3 AND len(images.shape)=4."
            )
        if not images_include_batch:
            images = ops.expand_dims(images, axis=0)

    if not boxes_includes_batch:
        return (
            ops.expand_dims(boxes, axis=0),
            ops.expand_dims(classes, axis=0),
            images,
            True,
        )
    return boxes, classes, images, False


def _format_outputs(boxes, classes, squeeze):
    if squeeze:
        return ops.squeeze(boxes, axis=0), ops.squeeze(classes, axis=0)
    return boxes, classes



# Copyright 2022 The KerasCV Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""VGG19 model for KerasCV.
Reference:
  - [Very Deep Convolutional Networks for Large-Scale Image Recognition](https://arxiv.org/abs/1409.1556)
    (ICLR 2015)
"""  # noqa: E501
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers

from keras_cv.src.models.legacy import utils
from keras_cv.src.models.legacy.vgg16 import apply_vgg_block


@keras.utils.register_keras_serializable(package="keras_cv.models")
class VGG19(keras.Model):
    """
    Reference:
    - [Very Deep Convolutional Networks for Large-Scale Image Recognition](https://arxiv.org/abs/1409.1556)
        (ICLR 2015)
    This class represents a Keras VGG19 model.
    Args:
      include_rescaling: bool, whether to rescale the inputs. If set to
        True, inputs will be passed through a `Rescaling(1/255.0)` layer.
      include_top: bool, whether to include the 3 fully-connected
        layers at the top of the network. If provided, num_classes must be
        provided.
      num_classes: int, optional number of classes to classify images into, only
        to be specified if `include_top` is True.
      weights: os.PathLike or None, one of `None` (random initialization), or a
        pretrained weight file path.
      input_shape: tuple, optional shape tuple, defaults to (224, 224, 3).
      input_tensor: Tensor, optional Keras tensor (i.e. output of
        `layers.Input()`) to use as image input for the model.
      pooling: bool, Optional pooling mode for feature extraction
        when `include_top` is `False`.
        - `None` means that the output of the model will be
            the 4D tensor output of the
            last convolutional block.
        - `avg` means that global average pooling
            will be applied to the output of the
            last convolutional block, and thus
            the output of the model will be a 2D tensor.
        - `max` means that global max pooling will
            be applied.
      classifier_activation:`str` or callable. The activation function to use
        on the "top" layer. Ignored unless `include_top=True`. Set
        `classifier_activation=None` to return the logits of the "top" layer.
        When loading pretrained weights, `classifier_activation` can only
        be `None` or `"softmax"`.
      name: (Optional) name to pass to the model, defaults to "VGG19".
    Returns:
      A `keras.Model` instance.
    """  # noqa: E501

    def __init__(
        self,
        include_rescaling,
        include_top,
        input_tensor=None,
        num_classes=None,
        weights=None,
        input_shape=(224, 224, 3),
        pooling=None,
        classifier_activation="softmax",
        name="VGG19",
        **kwargs,
    ):
        if weights and not tf.io.gfile.exists(weights):
            raise ValueError(
                "The `weights` argument should be either `None` or the path "
                "to the weights file to be loaded. Weights file not found at "
                "location: {weights}"
            )

        if include_top and not num_classes:
            raise ValueError(
                "If `include_top` is True, you should specify `num_classes`. "
                f"Received: num_classes={num_classes}"
            )

        if include_top and pooling:
            raise ValueError(
                f"`pooling` must be `None` when `include_top=True`."
                f"Received pooling={pooling} and include_top={include_top}. "
            )

        inputs = utils.parse_model_inputs(input_shape, input_tensor)
        x = inputs

        if include_rescaling:
            x = layers.Rescaling(1 / 255.0)(x)

        x = apply_vgg_block(
            x=x,
            num_layers=2,
            filters=64,
            kernel_size=(3, 3),
            activation="relu",
            padding="same",
            max_pool=True,
            name="block1",
        )

        x = apply_vgg_block(
            x=x,
            num_layers=2,
            filters=128,
            kernel_size=(3, 3),
            activation="relu",
            padding="same",
            max_pool=True,
            name="block2",
        )

        x = apply_vgg_block(
            x=x,
            num_layers=4,
            filters=256,
            kernel_size=(3, 3),
            activation="relu",
            padding="same",
            max_pool=True,
            name="block3",
        )

        x = apply_vgg_block(
            x=x,
            num_layers=4,
            filters=512,
            kernel_size=(3, 3),
            activation="relu",
            padding="same",
            max_pool=True,
            name="block4",
        )

        x = apply_vgg_block(
            x=x,
            num_layers=4,
            filters=512,
            kernel_size=(3, 3),
            activation="relu",
            padding="same",
            max_pool=True,
            name="block5",
        )

        if include_top:
            x = layers.Flatten(name="flatten")(x)
            x = layers.Dense(4096, activation="relu", name="fc1")(x)
            x = layers.Dense(4096, activation="relu", name="fc2")(x)
            x = layers.Dense(
                num_classes,
                activation=classifier_activation,
                name="predictions",
            )(x)
        else:
            if pooling == "avg":
                x = layers.GlobalAveragePooling2D()(x)
            elif pooling == "max":
                x = layers.GlobalMaxPooling2D()(x)

        super().__init__(inputs=inputs, outputs=x, name=name, **kwargs)
        if weights is not None:
            self.load_weights(weights)

        self.include_rescaling = include_rescaling
        self.include_top = include_top
        self.num_classes = num_classes
        self.input_tensor = input_tensor
        self.pooling = pooling
        self.classifier_activation = classifier_activation

    def get_config(self):
        return {
            "include_rescaling": self.include_rescaling,
            "include_top": self.include_top,
            "name": self.name,
            "input_shape": self.input_shape[1:],
            "input_tensor": self.input_tensor,
            "pooling": self.pooling,
            "num_classes": self.num_classes,
            "classifier_activation": self.classifier_activation,
            "trainable": self.trainable,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.layers.object_detection_3d import voxel_utils


def decode_bin_heading(predictions, num_bin):
    """Decode bin heading.

    Computes the box heading (orientation) by decoding the bin predictions. The
    predictions should contain bin classification scores (first num_bin scores)
    and corresponding bin residuals (the following num_bin scores).

    Args:
      predictions: Prediction scores tensor with size [N, num_bin*2]
        predictions = [:, bin_1, bin_2, ..., bin_k, res_1, res_2, ..., res_k],
        where k is the number of bins and N is the number of boxes.
      num_bin: A constant showing the number of bins used in heading bin loss.

    Returns:
      heading: Decoded heading tensor with size [N] in which heading values are
        in the [-pi, pi] range.

    Raises:
      ValueError: If the rank of `predictions` is not 2 or `predictions` tensor
        does not more than the expected number of dimensions.
    """
    with keras.backend.name_scope("decode_bin_heading"):
        if len(predictions.shape) != 2:
            raise ValueError(
                "The rank of the prediction tensor is expected to be 2. "
                f"Instead it is : {len(predictions.shape)}."
            )

        # Get the index of the bin with the maximum score to build a tensor of
        # [N].
        bin_idx = ops.cast(
            ops.argmax(predictions[:, 0:num_bin], axis=-1), "int32"
        )
        bin_idx_float = ops.cast(bin_idx, dtype=predictions.dtype)
        residual_norm = ops.take_along_axis(
            predictions[:, num_bin : num_bin * 2],
            ops.expand_dims(bin_idx, axis=-1),
            axis=-1,
        )[:, 0]

        # Divide 2pi into equal sized bins to compute the angle per class/bin.
        angle_per_class = (2 * np.pi) / num_bin
        residual_angle = residual_norm * (angle_per_class / 2)

        # bin_center is computed using the bin_idx and angle_per class,
        # (e.g., 0, 30, 60, 90, 120, ..., 270, 300, 330). Then residual is
        # added.
        heading = ops.mod(
            bin_idx_float * angle_per_class + residual_angle, 2 * np.pi
        )
        heading_mask = heading > np.pi
        heading = ops.where(heading_mask, heading - 2 * np.pi, heading)
        return heading


def decode_bin_box(pd, num_head_bin, anchor_size):
    """Decode bin based box encoding."""
    with keras.backend.name_scope("decode_bin_box"):
        delta = []
        start = 0
        for dim in [0, 1, 2]:
            delta.append(pd[:, start])
            start = start + 1

        heading = decode_bin_heading(pd[:, start:], num_head_bin)
        start = start + num_head_bin * 2

        size_res_norm = pd[:, start : start + 3]
        # [N,3]
        lwh = ops.cast(
            size_res_norm
            * ops.array(list(anchor_size), dtype=size_res_norm.dtype)
            + ops.array(list(anchor_size), dtype=size_res_norm.dtype),
            pd.dtype,
        )

        loc = ops.stack(delta, axis=-1)
        box = ops.concatenate(
            [loc, lwh, ops.expand_dims(heading, axis=-1)], axis=-1
        )
        return box


@keras_cv_export("keras_cv.layers.HeatmapDecoder")
class HeatmapDecoder(keras.layers.Layer):
    """A Keras layer that decodes predictions of a 3d object detection model.

    Arg:
      class_id: the integer index for a particular class.
      num_head_bin: number of bin classes divided by [-2pi, 2pi].
      anchor_size: the size of anchor at each xyz dimension.
      max_pool_size: the 2d pooling size for heatmap.
      max_num_box: top number of boxes select from heatmap.
      heatmap_threshold: the threshold to set a heatmap as positive.
      voxel_size: the x, y, z dimension of each voxel.
      spatial_size: the x, y, z boundary of voxels.
    """

    def __init__(
        self,
        class_id,
        num_head_bin,
        anchor_size,
        max_pool_size,
        max_num_box,
        heatmap_threshold,
        voxel_size,
        spatial_size,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.class_id = class_id
        self.num_head_bin = num_head_bin
        self.anchor_size = anchor_size
        self.max_pool_size = max_pool_size
        self.max_num_box = max_num_box
        self.heatmap_threshold = heatmap_threshold
        self.voxel_size = voxel_size
        self.spatial_size = spatial_size
        self.built = True

    def call(self, prediction):
        """Accepts raw predictions, and returns decoded boxes.

        Args:
            prediction: float Tensor.
        """
        heatmap = ops.softmax(prediction[..., :2])[..., 1:2]
        heatmap_pool = ops.max_pool(heatmap, self.max_pool_size, 1, "same")
        heatmap_mask = heatmap > self.heatmap_threshold
        heatmap_local_maxima_mask = ops.equal(heatmap, heatmap_pool)
        # [B, H, W, 1]
        heatmap_mask = ops.logical_and(heatmap_mask, heatmap_local_maxima_mask)
        # [B, H, W, 1]
        heatmap = ops.where(heatmap_mask, heatmap, 0)
        # [B, H, W]
        heatmap = ops.squeeze(heatmap, axis=-1)

        b, h, w = ops.shape(heatmap)
        heatmap = ops.reshape(heatmap, [b, h * w])
        _, top_index = ops.top_k(heatmap, k=self.max_num_box)

        # [B, H, W, ?]
        box_prediction = prediction[:, :, :, 2:]
        f = box_prediction.shape[-1]
        box_prediction = ops.reshape(box_prediction, [b, h * w, f])
        heatmap = ops.reshape(heatmap, [b, h * w])
        # [B, max_num_box, ?]
        box_prediction = ops.take_along_axis(
            box_prediction, ops.expand_dims(top_index, axis=-1), axis=1
        )
        # [B, max_num_box]
        box_score = ops.take_along_axis(heatmap, top_index, axis=1)
        box_class = ops.ones_like(box_score, "int32") * self.class_id
        # [B*max_num_box, ?]
        f = ops.shape(box_prediction)[-1]
        box_prediction_reshape = ops.reshape(
            box_prediction, [b * self.max_num_box, f]
        )
        # [B*max_num_box, 7]
        box_decoded = decode_bin_box(
            box_prediction_reshape, self.num_head_bin, self.anchor_size
        )
        # [B, max_num_box, 7]
        box_decoded = ops.reshape(box_decoded, [b, self.max_num_box, 7])
        global_xyz = ops.zeros([b, 3])
        ref_xyz = voxel_utils.compute_feature_map_ref_xyz(
            self.voxel_size, self.spatial_size, global_xyz
        )
        # [B, H, W, 3]
        ref_xyz = ops.squeeze(ref_xyz, axis=-2)
        f = list(ref_xyz.shape)[-1]
        ref_xyz = ops.reshape(ref_xyz, [b, h * w, f])
        # [B, max_num_box, 3]
        ref_xyz = ops.take_along_axis(
            ref_xyz, ops.expand_dims(top_index, axis=-1), axis=1
        )

        box_decoded_cxyz = ops.cast(
            ref_xyz + box_decoded[:, :, :3], box_decoded.dtype
        )
        box_decoded = ops.concatenate(
            [box_decoded_cxyz, box_decoded[:, :, 3:]], axis=-1
        )
        return box_decoded, box_class, box_score

    def get_config(self):
        config = {
            "class_id": self.class_id,
            "num_head_bin": self.num_head_bin,
            "anchor_size": self.anchor_size,
            "max_pool_size": self.max_pool_size,
            "max_num_box": self.max_num_box,
            "heatmap_threshold": self.heatmap_threshold,
            "voxel_size": self.voxel_size,
            "spatial_size": self.spatial_size,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from keras_cv.src import bounding_box
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


def _feature_bilinear_interpolation(features, kernel_y, kernel_x):
    """
    Feature bilinear interpolation.

    The RoIAlign feature f can be computed by bilinear interpolation
    of four neighboring feature points f0, f1, f2, and f3.
    f(y, x) = [hy, ly] * [[f00, f01], * [hx, lx]^T
                          [f10, f11]]
    f(y, x) = (hy*hx)f00 + (hy*lx)f01 + (ly*hx)f10 + (lx*ly)f11
    f(y, x) = w00*f00 + w01*f01 + w10*f10 + w11*f11
    kernel_y = [hy, ly]
    kernel_x = [hx, lx]

    Args:
      features: The features are in shape of [batch_size, num_boxes,
        output_size * 2, output_size * 2, num_filters].
      kernel_y: Tensor of size [batch_size, boxes, output_size, 2, 1].
      kernel_x: Tensor of size [batch_size, boxes, output_size, 2, 1].

    Returns:
      A 5-D tensor representing feature crop of shape
      [batch_size, num_boxes, output_size, output_size, num_filters].
    """
    features_shape = ops.shape(features)
    batch_size, num_boxes, output_size, num_filters = (
        features_shape[0],
        features_shape[1],
        features_shape[2],
        features_shape[4],
    )

    output_size = output_size // 2
    kernel_y = ops.reshape(
        kernel_y, [batch_size, num_boxes, output_size * 2, 1]
    )
    kernel_x = ops.reshape(
        kernel_x, [batch_size, num_boxes, 1, output_size * 2]
    )
    # Use implicit broadcast to generate the interpolation kernel. The
    # multiplier `4` is for avg pooling.
    interpolation_kernel = kernel_y * kernel_x * 4

    # Interpolate the gathered features with computed interpolation kernels.
    features *= ops.cast(
        ops.expand_dims(interpolation_kernel, axis=-1), dtype=features.dtype
    )
    features = ops.reshape(
        features,
        [batch_size * num_boxes, output_size * 2, output_size * 2, num_filters],
    )
    features = ops.nn.average_pool(
        features, [1, 2, 2, 1], [1, 2, 2, 1], "VALID"
    )
    features = ops.reshape(
        features, [batch_size, num_boxes, output_size, output_size, num_filters]
    )
    return features


def _compute_grid_positions(
    boxes,
    boundaries,
    output_size,
    sample_offset,
):
    """
    Computes the grid position w.r.t. the corresponding feature map.

    Args:
      boxes: a 3-D tensor of shape [batch_size, num_boxes, 4] encoding the
        information of each box w.r.t. the corresponding feature map.
        boxes[:, :, 0:2] are the grid position in (y, x) (float) of the top-left
        corner of each box. boxes[:, :, 2:4] are the box sizes in (h, w) (float)
          in terms of the number of pixels of the corresponding feature map
          size.
      boundaries: a 3-D tensor of shape [batch_size, num_boxes, 2] representing
        the boundary (in (y, x)) of the corresponding feature map for each box.
        Any resampled grid points that go beyond the boundary will be clipped.
      output_size: a scalar indicating the output crop size.
      sample_offset: a float number in [0, 1] indicates the subpixel sample
        offset from grid point.

    Returns:
      kernel_y: Tensor of size [batch_size, boxes, output_size, 2, 1].
      kernel_x: Tensor of size [batch_size, boxes, output_size, 2, 1].
      box_grid_y0y1: Tensor of size [batch_size, boxes, output_size, 2]
      box_grid_x0x1: Tensor of size [batch_size, boxes, output_size, 2]
    """
    boxes_shape = ops.shape(boxes)
    batch_size, num_boxes = boxes_shape[0], boxes_shape[1]
    if batch_size is None:
        batch_size = ops.shape(boxes)[0]
    box_grid_x = []
    box_grid_y = []
    for i in range(output_size):
        box_grid_x.append(
            boxes[:, :, 1] + (i + sample_offset) * boxes[:, :, 3] / output_size
        )
        box_grid_y.append(
            boxes[:, :, 0] + (i + sample_offset) * boxes[:, :, 2] / output_size
        )
    box_grid_x = ops.stack(box_grid_x, axis=2)
    box_grid_y = ops.stack(box_grid_y, axis=2)

    box_grid_y0 = ops.floor(box_grid_y)
    box_grid_x0 = ops.floor(box_grid_x)
    box_grid_x0 = ops.maximum(
        ops.cast(0.0, dtype=box_grid_x0.dtype), box_grid_x0
    )
    box_grid_y0 = ops.maximum(
        ops.cast(0.0, dtype=box_grid_y0.dtype), box_grid_y0
    )

    box_grid_x0 = ops.minimum(
        box_grid_x0, ops.expand_dims(boundaries[:, :, 1], -1)
    )
    box_grid_x1 = ops.minimum(
        box_grid_x0 + 1, ops.expand_dims(boundaries[:, :, 1], -1)
    )
    box_grid_y0 = ops.minimum(
        box_grid_y0, ops.expand_dims(boundaries[:, :, 0], -1)
    )
    box_grid_y1 = ops.minimum(
        box_grid_y0 + 1, ops.expand_dims(boundaries[:, :, 0], -1)
    )

    box_gridx0x1 = ops.stack([box_grid_x0, box_grid_x1], axis=-1)
    box_gridy0y1 = ops.stack([box_grid_y0, box_grid_y1], axis=-1)

    # The RoIAlign feature f can be computed by bilinear interpolation of four
    # neighboring feature points f0, f1, f2, and f3.
    # f(y, x) = [hy, ly] * [[f00, f01], * [hx, lx]^T
    #                       [f10, f11]]
    # f(y, x) = (hy*hx)f00 + (hy*lx)f01 + (ly*hx)f10 + (lx*ly)f11
    # f(y, x) = w00*f00 + w01*f01 + w10*f10 + w11*f11
    ly = box_grid_y - box_grid_y0
    lx = box_grid_x - box_grid_x0
    hy = 1.0 - ly
    hx = 1.0 - lx
    kernel_y = ops.reshape(
        ops.stack([hy, ly], axis=3), [batch_size, num_boxes, output_size, 2, 1]
    )
    kernel_x = ops.reshape(
        ops.stack([hx, lx], axis=3), [batch_size, num_boxes, output_size, 2, 1]
    )
    return kernel_y, kernel_x, box_gridy0y1, box_gridx0x1


def multilevel_crop_and_resize(
    features,
    boxes,
    output_size: int = 7,
    sample_offset: float = 0.5,
):
    """
    Crop and resize on multilevel feature pyramid.

    Generate the (output_size, output_size) set of pixels for each input box
    by first locating the box into the correct feature level, and then cropping
    and resizing it using the corresponding feature map of that level.

    Args:
      features: A dictionary with key as pyramid level and value as features.
        The pyramid level keys need to be represented by strings like so:
        "P2", "P3", "P4", and so on.
        The features are in shape of [batch_size, height_l, width_l,
        num_filters].
      boxes: A 3-D Tensor of shape [batch_size, num_boxes, 4]. Each row
        represents a box with [y1, x1, y2, x2] in un-normalized coordinates.
      output_size: A scalar to indicate the output crop size.
      sample_offset: a float number in [0, 1] indicates the subpixel sample
        offset from grid point.

    Returns:
      A 5-D tensor representing feature crop of shape
      [batch_size, num_boxes, output_size, output_size, num_filters].
    """

    levels_str = list(features.keys())
    # Levels are represented by strings with a prefix "P" to represent
    # pyramid levels. The integer level can be obtained by looking at
    # the value that follows the "P".
    levels = [int(level_str[1:]) for level_str in levels_str]
    min_level = min(levels)
    max_level = max(levels)
    features_shape = ops.shape(features[f"P{min_level}"])
    batch_size, max_feature_height, max_feature_width, num_filters = (
        features_shape[0],
        features_shape[1],
        features_shape[2],
        features_shape[3],
    )

    num_boxes = ops.shape(boxes)[1]

    # Stack feature pyramid into a features_all of shape
    # [batch_size, levels, height, width, num_filters].
    features_all = []
    feature_heights = []
    feature_widths = []
    for level in range(min_level, max_level + 1):
        shape = ops.shape(features[f"P{level}"])
        feature_heights.append(shape[1])
        feature_widths.append(shape[2])
        # Concat tensor of [batch_size, height_l * width_l, num_filters] for
        # each level.
        features_all.append(
            ops.reshape(features[f"P{level}"], [batch_size, -1, num_filters])
        )
    features_r2 = ops.reshape(
        ops.concatenate(features_all, 1), [-1, num_filters]
    )

    # Calculate height_l * width_l for each level.
    level_dim_sizes = [
        feature_widths[i] * feature_heights[i]
        for i in range(len(feature_widths))
    ]
    # level_dim_offsets is accumulated sum of level_dim_size.
    level_dim_offsets = [0]
    for i in range(len(feature_widths) - 1):
        level_dim_offsets.append(level_dim_offsets[i] + level_dim_sizes[i])
    batch_dim_size = level_dim_offsets[-1] + level_dim_sizes[-1]
    level_dim_offsets = (
        ops.ones_like(level_dim_offsets, dtype="int32") * level_dim_offsets
    )
    height_dim_sizes = (
        ops.ones_like(feature_widths, dtype="int32") * feature_widths
    )

    # Assigns boxes to the right level.
    box_width = boxes[:, :, 3] - boxes[:, :, 1]
    box_height = boxes[:, :, 2] - boxes[:, :, 0]
    areas_sqrt = ops.sqrt(
        ops.cast(box_height, "float32") * ops.cast(box_width, "float32")
    )

    # following the FPN paper to divide by 224.
    levels = ops.cast(
        ops.floor_divide(
            ops.log(ops.divide(areas_sqrt, 224.0)),
            ops.log(2.0),
        )
        + 4.0,
        dtype="int32",
    )
    # Maps levels between [min_level, max_level].
    levels = ops.minimum(max_level, ops.maximum(levels, min_level))

    # Projects box location and sizes to corresponding feature levels.
    scale_to_level = ops.cast(
        ops.power(2.0, ops.cast(levels, "float32")),
        dtype=boxes.dtype,
    )
    boxes /= ops.expand_dims(scale_to_level, axis=2)
    box_width /= scale_to_level
    box_height /= scale_to_level
    boxes = ops.concatenate(
        [
            boxes[:, :, 0:2],
            ops.expand_dims(box_height, -1),
            ops.expand_dims(box_width, -1),
        ],
        axis=-1,
    )

    # Maps levels to [0, max_level-min_level].
    levels -= min_level
    level_strides = ops.power([[2.0]], ops.cast(levels, "float32"))
    boundary = ops.cast(
        ops.concatenate(
            [
                ops.expand_dims(
                    [[ops.cast(max_feature_height, "float32")]] / level_strides
                    - 1,
                    axis=-1,
                ),
                ops.expand_dims(
                    [[ops.cast(max_feature_width, "float32")]] / level_strides
                    - 1,
                    axis=-1,
                ),
            ],
            axis=-1,
        ),
        boxes.dtype,
    )

    # Compute grid positions.
    (
        kernel_y,
        kernel_x,
        box_gridy0y1,
        box_gridx0x1,
    ) = _compute_grid_positions(boxes, boundary, output_size, sample_offset)

    x_indices = ops.cast(
        ops.reshape(box_gridx0x1, [batch_size, num_boxes, output_size * 2]),
        dtype="int32",
    )
    y_indices = ops.cast(
        ops.reshape(box_gridy0y1, [batch_size, num_boxes, output_size * 2]),
        dtype="int32",
    )

    batch_size_offset = ops.tile(
        ops.reshape(
            ops.arange(batch_size) * batch_dim_size, [batch_size, 1, 1, 1]
        ),
        [1, num_boxes, output_size * 2, output_size * 2],
    )
    # Get level offset for each box. Each box belongs to one level.
    levels_offset = ops.tile(
        ops.reshape(
            ops.take(level_dim_offsets, levels),
            [batch_size, num_boxes, 1, 1],
        ),
        [1, 1, output_size * 2, output_size * 2],
    )
    y_indices_offset = ops.tile(
        ops.reshape(
            y_indices * ops.expand_dims(ops.take(height_dim_sizes, levels), -1),
            [batch_size, num_boxes, output_size * 2, 1],
        ),
        [1, 1, 1, output_size * 2],
    )
    x_indices_offset = ops.tile(
        ops.reshape(x_indices, [batch_size, num_boxes, 1, output_size * 2]),
        [1, 1, output_size * 2, 1],
    )
    indices = ops.reshape(
        batch_size_offset + levels_offset + y_indices_offset + x_indices_offset,
        [-1],
    )

    # TODO(tanzhenyu): replace tf.gather with tf.gather_nd and try to get
    #  similar performance.
    features_per_box = ops.reshape(
        ops.take(features_r2, indices),
        [
            batch_size,
            num_boxes,
            output_size * 2,
            output_size * 2,
            num_filters,
        ],
    )

    # Bilinear interpolation.
    features_per_box = _feature_bilinear_interpolation(
        features_per_box, kernel_y, kernel_x
    )
    return features_per_box


# TODO(tanzhenyu): Remove this implementation once roi_pool has better
#  performance as this is mostly a duplicate of
#  https://github.com/tensorflow/models/blob/master/official/legacy/detection/ops/spatial_transform_ops.py#L324
@keras.utils.register_keras_serializable(package="keras_cv")
class _ROIAligner(keras.layers.Layer):
    """Performs ROIAlign for the second stage processing."""

    def __init__(
        self,
        bounding_box_format,
        target_size=7,
        sample_offset: float = 0.5,
        **kwargs,
    ):
        """
        Generates ROI Aligner.

        Args:
          bounding_box_format: the input format for boxes.
          crop_size: An `int` of the output size of the cropped features.
          sample_offset: A `float` in [0, 1] of the subpixel sample offset.
          **kwargs: Additional keyword arguments passed to Layer.
        """
        # assert_tf_keras("keras_cv.layers._ROIAligner")
        self._config_dict = {
            "bounding_box_format": bounding_box_format,
            "crop_size": target_size,
            "sample_offset": sample_offset,
        }
        super().__init__(**kwargs)

    def call(
        self,
        features,
        boxes,
        training: Optional[bool] = None,
    ):
        """

        Args:
          features: A dictionary with key as pyramid level and value as
            features. The features are in shape of
            [batch_size, height_l, width_l, num_filters].
          boxes: A 3-D `tf.Tensor` of shape [batch_size, num_boxes, 4]. Each row
            represents a box with [y1, x1, y2, x2] in un-normalized coordinates.
            from grid point.
          training: A `bool` of whether it is in training mode.
        Returns:
          A 5-D `tf.Tensor` representing feature crop of shape
          [batch_size, num_boxes, crop_size, crop_size, num_filters].
        """
        boxes = bounding_box.convert_format(
            boxes,
            source=self._config_dict["bounding_box_format"],
            target="yxyx",
        )
        roi_features = multilevel_crop_and_resize(
            features,
            boxes,
            output_size=self._config_dict["crop_size"],
            sample_offset=self._config_dict["sample_offset"],
        )
        return roi_features

    def get_config(self):
        return self._config_dict



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.bounding_box.converters import convert_format
from keras_cv.src.bounding_box.ensure_tensor import ensure_tensor
from keras_cv.src.bounding_box.formats import CENTER_XYWH
from keras_cv.src.bounding_box.formats import REL_XYWH
from keras_cv.src.bounding_box.formats import REL_XYXY
from keras_cv.src.bounding_box.formats import REL_YXYX
from keras_cv.src.bounding_box.formats import XYWH
from keras_cv.src.bounding_box.formats import XYXY
from keras_cv.src.bounding_box.formats import YXYX
from keras_cv.src.bounding_box.iou import compute_ciou
from keras_cv.src.bounding_box.iou import compute_iou
from keras_cv.src.bounding_box.mask_invalid_detections import (
    mask_invalid_detections,
)
from keras_cv.src.bounding_box.to_dense import to_dense
from keras_cv.src.bounding_box.to_ragged import to_ragged
from keras_cv.src.bounding_box.utils import as_relative
from keras_cv.src.bounding_box.utils import clip_to_image
from keras_cv.src.bounding_box.utils import is_relative
from keras_cv.src.bounding_box.validate_format import validate_format



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transformer to convert Waymo Open Dataset proto to model inputs."""

from typing import Any
from typing import Dict
from typing import List
from typing import Sequence
from typing import Tuple

import numpy as np
import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.utils import assert_waymo_open_dataset_installed

try:
    from waymo_open_dataset import dataset_pb2
    from waymo_open_dataset.utils import box_utils
    from waymo_open_dataset.utils import frame_utils
    from waymo_open_dataset.utils import range_image_utils
    from waymo_open_dataset.utils import transform_utils
except ImportError:
    waymo_open_dataset = None

from keras_cv.src.datasets.waymo import struct
from keras_cv.src.layers.object_detection_3d import voxel_utils

WOD_FRAME_OUTPUT_SIGNATURE = {
    "frame_id": tf.TensorSpec((), tf.int64),
    "timestamp_offset": tf.TensorSpec((), tf.float32),
    "timestamp_micros": tf.TensorSpec((), tf.int64),
    "pose": tf.TensorSpec([4, 4], tf.float32),
    "point_xyz": tf.TensorSpec([None, 3], tf.float32),
    "point_feature": tf.TensorSpec([None, 4], tf.float32),
    "point_mask": tf.TensorSpec([None], tf.bool),
    "point_range_image_row_col_sensor_id": tf.TensorSpec([None, 3], tf.float32),
    # Please refer to Waymo Open Dataset label proto for definitions.
    "label_box": tf.TensorSpec([None, 7], tf.float32),
    "label_box_id": tf.TensorSpec([None], tf.int64),
    "label_box_meta": tf.TensorSpec([None, 4], tf.float32),
    "label_box_class": tf.TensorSpec([None], tf.int32),
    "label_box_density": tf.TensorSpec([None], tf.int32),
    "label_box_detection_difficulty": tf.TensorSpec([None], tf.int32),
    "label_box_mask": tf.TensorSpec([None], tf.bool),
    "label_point_class": tf.TensorSpec([None], tf.int32),
    "label_point_nlz": tf.TensorSpec([None], tf.int32),
}

# Maximum number of points from all lidars excluding the top lidar. Please refer
# to https://arxiv.org/pdf/1912.04838.pdf Figure 1 for sensor layouts.
_MAX_NUM_NON_TOP_LIDAR_POINTS = 30000


def _decode_range_images(frame) -> Dict[int, List[tf.Tensor]]:
    """Decodes range images from a Waymo Open Dataset frame.

    Please refer to https://arxiv.org/pdf/1912.04838.pdf for more details.

    Args:
      frame: a Waymo Open Dataset frame.

    Returns:
      A dictionary mapping from sensor ID to list of range images ordered by
        return indices.
    """
    range_images = {}
    for lidar in frame.lasers:
        range_image_str_tensor = tf.io.decode_compressed(
            lidar.ri_return1.range_image_compressed, "ZLIB"
        )
        ri = dataset_pb2.MatrixFloat()
        ri.ParseFromString(bytearray(range_image_str_tensor.numpy()))
        ri_tensor = tf.reshape(
            tf.convert_to_tensor(value=ri.data, dtype=tf.float32), ri.shape.dims
        )
        range_images[lidar.name] = [ri_tensor]

        if lidar.name == dataset_pb2.LaserName.TOP:
            range_image_str_tensor = tf.io.decode_compressed(
                lidar.ri_return2.range_image_compressed, "ZLIB"
            )
            ri = dataset_pb2.MatrixFloat()
            ri.ParseFromString(bytearray(range_image_str_tensor.numpy()))
            ri_tensor = tf.reshape(
                tf.convert_to_tensor(value=ri.data, dtype=tf.float32),
                ri.shape.dims,
            )
            range_images[lidar.name].append(ri_tensor)
    return range_images


def _get_range_image_top_pose(frame) -> tf.Tensor:
    """Extracts range image pose tensor.

    Args:
      frame: a Waymo Open Dataset frame.

    Returns:
      Pose tensors for the range image.
    """
    _, _, _, ri_pose = frame_utils.parse_range_image_and_camera_projection(
        frame
    )
    assert ri_pose
    ri_pose_tensor = tf.reshape(
        tf.convert_to_tensor(value=ri_pose.data), ri_pose.shape.dims
    )
    # [H, W, 3, 3]
    ri_pose_tensor_rotation = transform_utils.get_rotation_matrix(
        ri_pose_tensor[..., 0], ri_pose_tensor[..., 1], ri_pose_tensor[..., 2]
    )
    ri_pose_tensor_translation = ri_pose_tensor[..., 3:]
    ri_pose_tensor = transform_utils.get_transform(
        ri_pose_tensor_rotation, ri_pose_tensor_translation
    )
    return ri_pose_tensor


def _get_point_top_lidar(
    range_image: Sequence[tf.Tensor], frame
) -> struct.PointTensors:
    """Gets point related tensors for the top lidar.

    Please refer to https://arxiv.org/pdf/1912.04838.pdf Table 2 for lidar
    specifications.

    Args:
      range_image: range image tensors. The range image is:
        [range, intensity, elongation, is_in_nlz].
      frame: a Waymo Open Dataset frame.

    Returns:
      Point tensors.
    """
    assert len(range_image) == 2
    xyz_list = []
    feature_list = []
    row_col_list = []
    nlz_list = []
    has_second_return_list = []
    is_second_return_list = []

    # Extracts frame pose tensor.
    frame_pose_tensor = tf.convert_to_tensor(
        value=np.reshape(np.array(frame.pose.transform), [4, 4])
    )

    # Extracts range image pose tensor.
    ri_pose_tensor = _get_range_image_top_pose(frame)

    # Extracts calibration data.
    calibration = _get_lidar_calibration(frame, dataset_pb2.LaserName.TOP)
    extrinsic = tf.reshape(np.array(calibration.extrinsic.transform), [4, 4])
    beam_inclinations = tf.constant(calibration.beam_inclinations)
    beam_inclinations = tf.reverse(beam_inclinations, axis=[-1])

    for i in range(2):
        ri_tensor = range_image[i]
        mask = ri_tensor[:, :, 0] > 0
        mask_idx = tf.cast(tf.where(mask), dtype=tf.int32)

        xyz = range_image_utils.extract_point_cloud_from_range_image(
            tf.expand_dims(ri_tensor[..., 0], axis=0),
            tf.expand_dims(extrinsic, axis=0),
            tf.expand_dims(beam_inclinations, axis=0),
            pixel_pose=tf.expand_dims(ri_pose_tensor, axis=0),
            frame_pose=tf.expand_dims(frame_pose_tensor, axis=0),
        )
        xyz = tf.gather_nd(tf.squeeze(xyz, axis=0), mask_idx)
        feature = tf.gather_nd(ri_tensor[:, :, 1:3], mask_idx)
        nlz = tf.gather_nd(ri_tensor[:, :, -1] > 0, mask_idx)

        xyz_list.append(xyz)
        feature_list.append(feature)
        nlz_list.append(nlz)
        row_col_list.append(mask_idx)

        if i == 0:
            has_second_return = range_image[1][:, :, 0] > 0
            has_second_return_list.append(
                tf.gather_nd(has_second_return, mask_idx)
            )
            is_second_return_list.append(
                tf.zeros([mask_idx.shape[0]], dtype=tf.bool)
            )
        else:
            has_second_return_list.append(
                tf.zeros([mask_idx.shape[0]], dtype=tf.bool)
            )
            is_second_return_list.append(
                tf.ones([mask_idx.shape[0]], dtype=tf.bool)
            )

    xyz = tf.concat(xyz_list, axis=0)
    feature = tf.concat(feature_list, axis=0)
    row_col = tf.concat(row_col_list, axis=0)
    nlz = tf.concat(nlz_list, axis=0)
    has_second_return = tf.cast(
        tf.concat(has_second_return_list, axis=0), dtype=tf.float32
    )
    is_second_return = tf.cast(
        tf.concat(is_second_return_list, axis=0), dtype=tf.float32
    )
    # Complete feature: intensity, elongation, has_second, is_second.
    feature = tf.concat(
        [
            feature,
            has_second_return[:, tf.newaxis],
            is_second_return[:, tf.newaxis],
        ],
        axis=-1,
    )
    sensor_id = (
        tf.ones([xyz.shape[0], 1], dtype=tf.int32) * dataset_pb2.LaserName.TOP
    )
    ri_row_col_sensor_id = tf.concat([row_col, sensor_id], axis=-1)

    return struct.PointTensors(
        point_xyz=xyz,
        point_feature=feature,
        point_range_image_row_col_sensor_id=ri_row_col_sensor_id,
        label_point_nlz=nlz,
    )


def _get_lidar_calibration(frame, name: int):
    """Gets lidar calibration for a given lidar."""
    calibration = None
    for c in frame.context.laser_calibrations:
        if c.name == name:
            calibration = c
    assert calibration is not None
    return calibration


def _downsample(point: struct.PointTensors, n: int) -> struct.PointTensors:
    """Randomly samples up to n points from the given point_tensor."""
    num_points = point.point_xyz.shape[0]
    if num_points <= n:
        return point
    mask = tf.range(start=0, limit=num_points, dtype=tf.int32)
    mask = tf.random.shuffle(mask)
    mask_index = mask[:n]

    def _gather(t: tf.Tensor) -> tf.Tensor:
        return tf.gather(t, mask_index)

    tensors = {key: _gather(value) for key, value in vars(point).items()}
    return struct.PointTensors(**tensors)


def _get_point_lidar(
    ris: Dict[int, List[tf.Tensor]],
    frame,
    max_num_points: int,
) -> struct.PointTensors:
    """Gets point related tensors for non-top lidar.

    The main differences from top lidar extraction are related to second return
    and point down sampling.

    Args:
      ris: Mapping from lidar ID to range image tensor. The ri format is [range,
        intensity, elongation, is_in_nlz].
      frame: a Waymo Open Dataset frame.
      max_num_points: maximum number of points from non-top lidar.

    Returns:
      Point related tensors.
    """
    xyz_list = []
    feature_list = []
    nlz_list = []
    ri_row_col_sensor_id_list = []

    for sensor_id in ris.keys():
        ri_tensor = ris[sensor_id]
        assert len(ri_tensor) == 1, f"{sensor_id}"
        ri_tensor = ri_tensor[0]
        calibration = _get_lidar_calibration(frame, sensor_id)
        extrinsic = tf.reshape(
            np.array(calibration.extrinsic.transform), [4, 4]
        )
        beam_inclinations = range_image_utils.compute_inclination(
            tf.constant(
                [
                    calibration.beam_inclination_min,
                    calibration.beam_inclination_max,
                ]
            ),
            height=ri_tensor.shape[0],
        )
        beam_inclinations = tf.reverse(beam_inclinations, axis=[-1])
        xyz = range_image_utils.extract_point_cloud_from_range_image(
            tf.expand_dims(ri_tensor[..., 0], axis=0),
            tf.expand_dims(extrinsic, axis=0),
            tf.expand_dims(beam_inclinations, axis=0),
        )
        mask = ri_tensor[:, :, 0] > 0
        mask_idx = tf.cast(tf.where(mask), dtype=tf.int32)

        xyz = tf.gather_nd(tf.squeeze(xyz, axis=0), mask_idx)
        feature = tf.gather_nd(ri_tensor[:, :, 1:3], mask_idx)
        feature = tf.concat(
            [feature, tf.zeros([feature.shape[0], 2], dtype=tf.float32)],
            axis=-1,
        )
        nlz = tf.gather_nd(ri_tensor[:, :, -1] > 0, mask_idx)

        xyz_list.append(xyz)
        feature_list.append(feature)
        nlz_list.append(nlz)
        ri_row_col_sensor_id_list.append(
            tf.concat(
                [
                    mask_idx,
                    sensor_id * tf.ones([nlz.shape[0], 1], dtype=tf.int32),
                ],
                axis=-1,
            )
        )

    xyz = tf.concat(xyz_list, axis=0)
    feature = tf.concat(feature_list, axis=0)
    nlz = tf.concat(nlz_list, axis=0)
    ri_row_col_sensor_id = tf.concat(ri_row_col_sensor_id_list, axis=0)

    point_tensors = struct.PointTensors(
        point_xyz=xyz,
        point_feature=feature,
        point_range_image_row_col_sensor_id=ri_row_col_sensor_id,
        label_point_nlz=nlz,
    )
    point_tensors = _downsample(point_tensors, max_num_points)
    return point_tensors


def _get_point(frame, max_num_lidar_points: int) -> struct.PointTensors:
    """Gets point related tensors from a Waymo Open Dataset frame.

    Args:
      frame: a Waymo Open Dataset frame.
      max_num_lidar_points: maximum number of points from non-top lidars.

    Returns:
      Point related tensors.
    """
    range_images = _decode_range_images(frame)
    point_top_lidar = _get_point_top_lidar(
        range_images[dataset_pb2.LaserName.TOP], frame
    )

    range_images.pop(dataset_pb2.LaserName.TOP)
    point_tensors_lidar = _get_point_lidar(
        range_images, frame, max_num_lidar_points
    )

    merged = {}
    for key in vars(point_tensors_lidar).keys():
        merged[key] = tf.concat(
            [getattr(point_tensors_lidar, key), getattr(point_top_lidar, key)],
            axis=0,
        )
    return struct.PointTensors(**merged)


def _get_point_label_box(
    frame,
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Extracts 3D box labels from a Waymo Open Dataset frame.

    Args:
      frame: a Waymo Open Dataset frame.

    Returns:
      box_3d: [M, 7] 3d boxes.
      box_meta: [M, 4] speed and accel for each box.
      box_class: [M] object class of each box.
      box_id: [M] unique ID of each box.
      box_density: [M] number of points in each box.
      box_detection_difficulty: [M] difficulty level for detection.
    """
    box_3d_list = []
    box_meta_list = []
    box_class_list = []
    box_id_list = []
    box_density_list = []
    box_detection_difficulty_list = []

    for label in frame.laser_labels:
        model_object_type = label.type
        density = label.num_lidar_points_in_box
        detection_difficulty = label.detection_difficulty_level

        if model_object_type == 0:
            continue
        b = label.box
        box_3d_list.extend(
            [
                b.center_x,
                b.center_y,
                b.center_z,
                b.length,
                b.width,
                b.height,
                b.heading,
            ]
        )
        meta = label.metadata
        box_meta_list.extend(
            [
                meta.speed_x,
                meta.speed_y,
                meta.accel_x,
                meta.accel_y,
            ]
        )
        box_class_list.append(model_object_type)
        box_id = tf.bitcast(
            tf.fingerprint(
                tf.expand_dims(label.id.encode(encoding="ascii"), 0)
            )[0],
            tf.int64,
        )
        box_id_list.append(box_id)
        box_density_list.append(density)
        box_detection_difficulty_list.append(detection_difficulty)

    box_3d = tf.reshape(tf.constant(box_3d_list, dtype=tf.float32), [-1, 7])
    box_meta = tf.reshape(tf.constant(box_meta_list, dtype=tf.float32), [-1, 4])
    box_class = tf.constant(box_class_list, dtype=tf.int32)
    box_id = tf.stack(box_id_list)
    box_density = tf.constant(box_density_list, dtype=tf.int32)
    box_detection_difficulty = tf.constant(
        box_detection_difficulty_list, dtype=tf.int32
    )
    return (
        box_3d,
        box_meta,
        box_class,
        box_id,
        box_density,
        box_detection_difficulty,
    )


def _get_box_class_per_point(
    box: tf.Tensor, box_class: tf.Tensor, point_xyz: tf.Tensor
) -> tf.Tensor:
    """Extracts point labels.

    Args:
      box: [M, 7] box tensor.
      box_class: [M] class of each box.
      point_xyz: [N, 3] points.

    Returns:
      point_box_class: [N] box class of each point.
    """
    n = point_xyz.shape[0]
    m = box.shape[0]
    if m == 0:
        return tf.zeros([n], dtype=tf.int32)

    # [N, M]
    point_in_box = box_utils.is_within_box_3d(point_xyz, box)
    # [N]
    point_in_any_box = tf.math.reduce_any(point_in_box, axis=-1)
    # [N]
    point_box_idx = tf.math.argmax(point_in_box, axis=-1, output_type=tf.int32)
    # [N]
    point_box_class = tf.where(
        point_in_any_box, tf.gather(box_class, point_box_idx), 0
    )

    return point_box_class


def _get_point_label(frame, point_xyz: tf.Tensor) -> struct.LabelTensors:
    """Extracts labels.

    Args:
      frame: an open dataset frame.
      point_xyz: [N, 3] tensor representing point xyz.

    Returns:
      Label tensors.
    """
    (
        box_3d,
        box_meta,
        box_class,
        box_id,
        box_density,
        box_detection_difficulty,
    ) = _get_point_label_box(frame)
    point_box_class = _get_box_class_per_point(box_3d, box_class, point_xyz)
    box_mask = tf.math.greater(box_class, 0)
    return struct.LabelTensors(
        label_box=box_3d,
        label_box_id=box_id,
        label_box_meta=box_meta,
        label_box_class=box_class,
        label_box_density=box_density,
        label_box_detection_difficulty=box_detection_difficulty,
        label_box_mask=box_mask,
        label_point_class=point_box_class,
    )


def _point_vehicle_to_global(
    point_vehicle_xyz: tf.Tensor, sdc_pose: tf.Tensor
) -> tf.Tensor:
    """Transforms points from vehicle to global frame.

    Args:
      point_vehicle_xyz: [..., N, 3] vehicle xyz.
      sdc_pose: [..., 4, 4] the SDC pose.

    Returns:
      The points in global frame.
    """
    rot = sdc_pose[..., 0:3, 0:3]
    loc = sdc_pose[..., 0:3, 3]
    return (
        tf.linalg.matmul(point_vehicle_xyz, rot, transpose_b=True)
        + loc[..., tf.newaxis, :]
    )


def _point_global_to_vehicle(
    point_xyz: tf.Tensor, sdc_pose: tf.Tensor
) -> tf.Tensor:
    """Transforms points from global to vehicle frame.

    Args:
      point_xyz: [..., N, 3] global xyz.
      sdc_pose: [..., 4, 4] the SDC pose.

    Returns:
      The points in vehicle frame.
    """
    rot = sdc_pose[..., 0:3, 0:3]
    loc = sdc_pose[..., 0:3, 3]
    return (
        tf.linalg.matmul(point_xyz, rot)
        + voxel_utils.inv_loc(rot, loc)[..., tf.newaxis, :]
    )


def _box_3d_vehicle_to_global(
    box_3d: tf.Tensor, sdc_pose: tf.Tensor
) -> tf.Tensor:
    """Transforms 3D boxes from vehicle to global frame.

    Args:
      box_3d: [..., N, 7] 3d boxes in vehicle frame.
      sdc_pose: [..., 4, 4] the SDC pose.

    Returns:
      The boxes in global frame.
    """
    center = box_3d[..., 0:3]
    dim = box_3d[..., 3:6]
    heading = box_3d[..., 6]

    new_center = _point_vehicle_to_global(center, sdc_pose)
    new_heading = (
        heading
        + tf.atan2(sdc_pose[..., 1, 0], sdc_pose[..., 0, 0])[..., tf.newaxis]
    )

    return tf.concat([new_center, dim, new_heading[..., tf.newaxis]], axis=-1)


def _box_3d_global_to_vehicle(
    box_3d: tf.Tensor, sdc_pose: tf.Tensor
) -> tf.Tensor:
    """Transforms 3D boxes from global to vehicle frame.

    Args:
      box_3d: [..., N, 7] 3d boxes in global frame.
      sdc_pose: [..., 4, 4] the SDC pose.

    Returns:
      The boxes in vehicle frame.
    """
    center = box_3d[..., 0:3]
    dim = box_3d[..., 3:6]
    heading = box_3d[..., 6]

    new_center = _point_global_to_vehicle(center, sdc_pose)
    new_heading = (
        heading
        + tf.atan2(sdc_pose[..., 0, 1], sdc_pose[..., 0, 0])[..., tf.newaxis]
    )

    return tf.concat([new_center, dim, new_heading[..., tf.newaxis]], axis=-1)


@keras_cv_export("keras_cv.datasets.waymo.build_tensors_from_wod_frame")
def build_tensors_from_wod_frame(frame) -> Dict[str, tf.Tensor]:
    """Builds tensors from a Waymo Open Dataset frame.

    This function is to convert range image to point cloud. User can also work
    with range image directly with frame_utils functions from
    waymo_open_dataset.

    Args:
      frame: a Waymo Open Dataset frame.

    Returns:
      Flat dictionary of tensors.
    """
    assert_waymo_open_dataset_installed(
        "keras_cv.datasets.waymo.build_tensors_from_wod_frame()"
    )

    frame_id_bytes = "{}_{}".format(
        frame.context.name, frame.timestamp_micros
    ).encode(encoding="ascii")
    frame_id = tf.bitcast(
        tf.fingerprint(tf.expand_dims(frame_id_bytes, 0))[0], tf.int64
    )

    timestamp_micros = tf.constant(frame.timestamp_micros, dtype=tf.int64)
    pose = tf.convert_to_tensor(
        value=np.reshape(np.array(frame.pose.transform), [4, 4]),
        dtype_hint=tf.float32,
    )

    point_tensors = _get_point(frame, _MAX_NUM_NON_TOP_LIDAR_POINTS)
    point_label_tensors = _get_point_label(frame, point_tensors.point_xyz)

    # Transforms lidar frames to global coordinates.
    point_tensors.point_xyz = _point_vehicle_to_global(
        point_tensors.point_xyz, pose
    )
    point_label_tensors.label_box = _box_3d_vehicle_to_global(
        point_label_tensors.label_box, pose
    )

    # Constructs final results.
    num_points = point_tensors.point_xyz.shape[0]
    return {
        "frame_id": frame_id,
        "timestamp_offset": tf.constant(0.0, dtype=tf.float32),
        "timestamp_micros": timestamp_micros,
        "pose": pose,
        "point_xyz": point_tensors.point_xyz,
        "point_feature": point_tensors.point_feature,
        "point_mask": tf.ones([num_points], dtype=tf.bool),
        "point_range_image_row_col_sensor_id": point_tensors.point_range_image_row_col_sensor_id,  # noqa: E501
        "label_box": point_label_tensors.label_box,
        "label_box_id": point_label_tensors.label_box_id,
        "label_box_meta": point_label_tensors.label_box_meta,
        "label_box_class": point_label_tensors.label_box_class,
        "label_box_density": point_label_tensors.label_box_density,
        "label_box_detection_difficulty": point_label_tensors.label_box_detection_difficulty,  # noqa: E501
        "label_box_mask": point_label_tensors.label_box_mask,
        "label_point_class": point_label_tensors.label_point_class,
        "label_point_nlz": point_tensors.label_point_nlz,
    }


@keras_cv_export("keras_cv.datasets.waymo.pad_or_trim_tensors")
def pad_or_trim_tensors(
    frame: Dict[str, tf.Tensor], max_num_point=199600, max_num_label_box=1000
) -> Dict[str, tf.Tensor]:
    """Pad or trim tensors from a frame to have uniform shapes.

    Args:
      frame: a dictionary of feature tensors from a Waymo Open Dataset frame.
      max_num_point: maximum number of lidar points to process.
      max_num_label_box: maximum number of label boxes to process.


    Returns:
      A dictionary of feature tensors with uniform shapes.
    """

    def _pad_fn(t: tf.Tensor, max_counts: int) -> tf.Tensor:
        shape = [max_counts] + t.shape.as_list()[1:]
        return voxel_utils._pad_or_trim_to(t, shape)

    point_tensor_keys = {
        "point_xyz",
        "point_feature",
        "point_range_image_row_col_sensor_id",
        "point_mask",
        "label_point_class",
        "label_point_nlz",
    }
    box_tensor_keys = {
        "label_box",
        "label_box_id",
        "label_box_meta",
        "label_box_class",
        "label_box_density",
        "label_box_detection_difficulty",
        "label_box_mask",
    }
    for key in point_tensor_keys:
        t = frame[key]
        if t is not None:
            frame[key] = _pad_fn(t, max_num_point)
    for key in box_tensor_keys:
        t = frame[key]
        if t is not None:
            frame[key] = _pad_fn(t, max_num_label_box)
    return frame


@keras_cv_export("keras_cv.datasets.waymo.transform_to_vehicle_frame")
def transform_to_vehicle_frame(
    frame: Dict[str, tf.Tensor]
) -> Dict[str, tf.Tensor]:
    """Transform tensors in a frame from global coordinates to vehicle
    coordinates.

    Args:
      frame: a dictionary of feature tensors from a Waymo Open Dataset frame in
        global frame.


    Returns:
      A dictionary of feature tensors in vehicle frame.
    """
    assert_waymo_open_dataset_installed(
        "keras_cv.datasets.waymo.transform_to_vehicle_frame()"
    )

    def _transform_to_vehicle_frame(
        point_global_xyz: tf.Tensor,
        point_mask: tf.Tensor,
        box_global: tf.Tensor,
        box_mask: tf.Tensor,
        sdc_pose: tf.Tensor,
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        point_vehicle_xyz = _point_global_to_vehicle(point_global_xyz, sdc_pose)
        point_vehicle_xyz = tf.where(
            point_mask[..., tf.newaxis], point_vehicle_xyz, 0.0
        )
        box_vehicle = _box_3d_global_to_vehicle(box_global, sdc_pose)
        box_vehicle = tf.where(box_mask[..., tf.newaxis], box_vehicle, 0.0)
        return point_vehicle_xyz, box_vehicle

    point_vehicle_xyz, box_vehicle = _transform_to_vehicle_frame(
        frame["point_xyz"],
        frame["point_mask"],
        frame["label_box"],
        frame["label_box_mask"],
        frame["pose"],
    )
    frame["point_xyz"] = point_vehicle_xyz
    frame["label_box"] = box_vehicle
    # Override pose as the points and boxes are in the vehicle frame.
    frame["pose"] = tf.eye(4)
    if frame["label_point_nlz"] is not None:
        frame["point_mask"] = tf.logical_and(
            frame["point_mask"],
            tf.logical_not(tf.cast(frame["label_point_nlz"], tf.bool)),
        )
    return frame


@keras_cv_export("keras_cv.datasets.waymo.convert_to_center_pillar_inputs")
def convert_to_center_pillar_inputs(
    frame: Dict[str, tf.Tensor]
) -> Dict[str, Any]:
    """Converts an input frame into CenterPillar input format.

    Args:
      frame: a dictionary of feature tensors from a Waymo Open Dataset frame

    Returns:
      A dictionary of two tensor dictionaries with keys "point_clouds"
      and "3d_boxes".
    """
    point_clouds = {
        "point_xyz": frame["point_xyz"],
        "point_feature": frame["point_feature"],
        "point_mask": frame["point_mask"],
    }
    boxes = {
        "boxes": frame["label_box"],
        "classes": frame["label_box_class"],
        "difficulty": frame["label_box_detection_difficulty"],
        "mask": frame["label_box_mask"],
    }
    y = {
        "point_clouds": point_clouds,
        "3d_boxes": boxes,
    }
    return y


@keras_cv_export("keras_cv.datasets.waymo.build_tensors_for_augmentation")
def build_tensors_for_augmentation(
    frame: Dict[str, tf.Tensor]
) -> Tuple[tf.Tensor, tf.Tensor]:
    """Builds tensors for data augmentation from an input frame.

    Args:
      frame: a dictionary of feature tensors from a Waymo Open Dataset frame

    Returns:
      A dictionary of two tensors with keys "point_clouds" and "bounding_boxes"
      and values which are tensors of shapes [num points, num features] and
      [num boxes, num features]).
    """
    assert_waymo_open_dataset_installed(
        "keras_cv.datasets.waymo.build_tensors_for_augmentation()"
    )
    point_cloud = tf.concat(
        [
            frame["point_xyz"][tf.newaxis, ...],
            frame["point_feature"][tf.newaxis, ...],
            tf.cast(frame["point_mask"], tf.float32)[tf.newaxis, :, tf.newaxis],
        ],
        axis=-1,
    )
    boxes = tf.concat(
        [
            frame["label_box"][tf.newaxis, :],
            tf.cast(frame["label_box_class"], tf.float32)[
                tf.newaxis, :, tf.newaxis
            ],
            tf.cast(frame["label_box_mask"], tf.float32)[
                tf.newaxis, :, tf.newaxis
            ],
            tf.cast(frame["label_box_density"], tf.float32)[
                tf.newaxis, :, tf.newaxis
            ],
            tf.cast(frame["label_box_detection_difficulty"], tf.float32)[
                tf.newaxis, :, tf.newaxis
            ],
        ],
        axis=-1,
    )
    return {
        "point_clouds": tf.squeeze(point_cloud, axis=0),
        "bounding_boxes": tf.squeeze(boxes, axis=0),
    }



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
import tensorflow as tf

from keras_cv.src import utils
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import ops
from keras_cv.src.utils import assert_matplotlib_installed

try:
    import matplotlib.pyplot as plt
except:
    plt = None


def _extract_image_batch(images, num_images, batch_size):
    def unpack_images(inputs):
        return inputs["image"]

    num_batches_required = math.ceil(num_images / batch_size)

    if isinstance(images, tf.data.Dataset):
        images = images.map(unpack_images)

        if batch_size == 1:
            images = images.ragged_batch(num_batches_required)
            sample = next(iter(images.take(1)))
        else:
            sample = next(iter(images.take(num_batches_required)))

        return sample

    else:
        if len(ops.shape(images)) != 4:
            raise ValueError(
                "`plot_images_gallery()` requires you to "
                "batch your `np.array` samples together."
            )
        else:
            num_samples = (
                num_images if num_images <= batch_size else num_batches_required
            )
            sample = images[:num_samples, ...]
    return sample


@keras_cv_export("keras_cv.visualization.plot_image_gallery")
def plot_image_gallery(
    images,
    value_range,
    scale=2,
    rows=None,
    cols=None,
    path=None,
    show=None,
    transparent=True,
    dpi=60,
    legend_handles=None,
):
    """Displays a gallery of images.

    Example:
    ```python
    train_ds = tfds.load(
        "cats_vs_dogs",
        split="train",
        with_info=False,
        shuffle_files=True,
    )

    keras_cv.visualization.plot_image_gallery(
        train_ds,
        value_range=(0, 255),
        scale=3,
    )
    ```

    ![example gallery](https://i.imgur.com/r0ndse0.png)

    Args:
        images: a Tensor, `tf.data.Dataset` or NumPy array containing images
            to show in the gallery. Note: If using a `tf.data.Dataset`,
            images should be present in the `FeaturesDict` under
            the key `image`.
        value_range: value range of the images. Common examples include
            `(0, 255)` and `(0, 1)`.
        scale: how large to scale the images in the gallery
        rows: (Optional) number of rows in the gallery to show.
            Required if inputs are unbatched.
        cols: (Optional) number of columns in the gallery to show.
            Required if inputs are unbatched.
        path: (Optional) path to save the resulting gallery to.
        show: (Optional) whether to show the gallery of images.
        transparent: (Optional) whether to give the image a transparent
            background, defaults to `True`.
        dpi: (Optional) the dpi to pass to matplotlib.savefig(), defaults to
            `60`.
        legend_handles: (Optional) matplotlib.patches List of legend handles.
            I.e. passing: `[patches.Patch(color='red', label='mylabel')]` will
            produce a legend with a single red patch and the label 'mylabel'.

    """
    assert_matplotlib_installed("plot_bounding_box_gallery")

    if path is not None and show:
        raise ValueError(
            "plot_gallery() expects either `path` to be set, or `show` "
            "to be true."
        )

    if isinstance(images, tf.data.Dataset):
        sample = next(iter(images.take(1)))
        batch_size = (
            sample["image"].shape[0] if len(sample["image"].shape) == 4 else 1
        )  # batch_size from within passed `tf.data.Dataset`
    else:
        batch_size = (
            ops.shape(images)[0] if len(ops.shape(images)) == 4 else 1
        )  # batch_size from np.array or single image

    rows = rows or int(math.ceil(math.sqrt(batch_size)))
    cols = cols or int(math.ceil(batch_size // rows))

    num_images = rows * cols
    images = _extract_image_batch(images, num_images, batch_size)

    # Generate subplots
    fig, axes = plt.subplots(
        nrows=rows,
        ncols=cols,
        figsize=(cols * scale, rows * scale),
        frameon=False,
        layout="tight",
        squeeze=True,
        sharex="row",
        sharey="col",
    )
    fig.subplots_adjust(wspace=0, hspace=0)

    if isinstance(axes, np.ndarray) and len(axes.shape) == 1:
        expand_axis = 0 if rows == 1 else -1
        axes = np.expand_dims(axes, expand_axis)

    if legend_handles is not None:
        fig.legend(handles=legend_handles, loc="lower center")

    # Perform image range transform
    images = utils.transform_value_range(
        images, original_range=value_range, target_range=(0, 255)
    )
    images = utils.to_numpy(images)

    for row in range(rows):
        for col in range(cols):
            index = row * cols + col
            current_axis = (
                axes[row, col] if isinstance(axes, np.ndarray) else axes
            )
            current_axis.imshow(images[index].astype("uint8"))
            current_axis.margins(x=0, y=0)
            current_axis.axis("off")

    if path is None and show is None:
        return fig

    if path is not None:
        plt.savefig(
            fname=path,
            pad_inches=0,
            bbox_inches="tight",
            transparent=transparent,
            dpi=dpi,
        )
        plt.close()
    elif show:
        plt.show()
        plt.close()



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf

from keras_cv.src.utils.resource_loader import LazySO

custom_ops = LazySO("custom_ops/_keras_cv_custom_ops.so")


# TODO(tanzhenyu): remove assumption of non overlapping boxes
def within_box3d_index(points, boxes):
    """Assign point to the box index that it belongs to.
    If no box contains the point, it will be assigned -1.
    This v2 function assumes that bounding boxes DO NOT overlap with each other.

    Args:
      points: [..., num_points, 3] float32 Tensor for 3d points in xyz format.
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].

    Returns:
      integer Tensor of shape [..., num_points] indicating which box index each
        point belongs to.
    """
    points = tf.convert_to_tensor(points)
    boxes = tf.convert_to_tensor(boxes)
    if points.shape.rank == 2 and boxes.shape.rank == 2:
        return custom_ops.ops.kcv_within_box(points, boxes)
    elif points.shape.rank == 3 and boxes.shape.rank == 3:
        num_samples = points.get_shape().as_list()[0]
        results = []
        for i in range(num_samples):
            results.append(
                custom_ops.ops.kcv_within_box(points[i], boxes[i])[
                    tf.newaxis, ...
                ]
            )
        return tf.concat(results, axis=0)
    else:
        raise ValueError(
            "is_within_box3d_v2 are expecting inputs point clouds and bounding "
            "boxes to be rank 2D (Point, Feature) or 3D (Frame, Point, Feature)"
            " tensors. Got shape: {} and {}".format(points.shape, boxes.shape)
        )


def group_points_by_boxes(points, boxes):
    """Checks if 3d points are within 3d bounding boxes.
    Currently only xyz format is supported.
    This function assumes that bounding boxes DO NOT overlap with each other.

    Args:
      points: [..., num_points, 3] float32 Tensor for 3d points in xyz format.
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].

    Returns:
      boolean Ragged Tensor of shape [..., num_boxes, ragged_points] for each
      box, all the point indices that belong to the box.

    """
    num_boxes = boxes.shape[-2] or tf.shape(boxes)[-2]
    # [..., num_points]
    box_indices = within_box3d_index(points, boxes)
    num_points = points.shape[-2] or tf.shape(points)[-2]
    point_indices = tf.range(num_points, dtype=tf.int32)

    def group_per_sample(box_index):
        point_mask = tf.math.greater_equal(box_index, 0)
        valid_point_indices = tf.boolean_mask(point_indices, point_mask)
        valid_box_index = tf.boolean_mask(box_index, point_mask)
        res = tf.ragged.stack_dynamic_partitions(
            valid_point_indices, valid_box_index, num_partitions=num_boxes
        )
        return res

    boxes_rank = len(boxes.shape)
    if boxes_rank == 2:
        return group_per_sample(box_indices)
    elif boxes_rank == 3:
        num_samples = boxes.get_shape().as_list()[0]
        res_list = []
        for i in range(num_samples):
            res_list.append(group_per_sample(box_indices[i]))
        return tf.ragged.stack(res_list)
    else:
        raise ValueError(
            f"Does not support box rank > 3, got boxes shape {boxes.shape}"
        )


# TODO(lengzhaoqi/tanzhenyu): compare the performance with v1
def is_within_any_box3d_v2(points, boxes, keepdims=False):
    """Checks if 3d points are within 3d bounding boxes.
    Currently only xyz format is supported.

    Args:
      points: [..., num_points, 3] float32 Tensor for 3d points in xyz format.
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].
      keepdims: boolean. If true, retains reduced dimensions with length 1.

    Returns:
      boolean Tensor of shape [..., num_points] indicating whether
      the point belongs to the box.

    """
    res = tf.greater_equal(within_box3d_index(points, boxes), 0)
    if keepdims:
        res = res[..., tf.newaxis]
    return res


def is_within_any_box3d_v3(points, boxes, keepdims=False):
    """Checks if 3d points are within 3d bounding boxes.
    Currently only xyz format is supported.

    Args:
      points: [..., num_points, 3] float32 Tensor for 3d points in xyz format.
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].
      keepdims: boolean. If true, retains reduced dimensions with length 1.

    Returns:
      boolean Tensor of shape [..., num_points] indicating whether
      the point belongs to the box.

    """
    res = custom_ops.ops.kcv_within_any_box(points, boxes)
    if keepdims:
        res = res[..., tf.newaxis]
    return res


def get_rank(tensor):
    return tensor.shape.ndims or tf.rank(tensor)


def wrap_angle_radians(angle_radians, min_val=-np.pi, max_val=np.pi):
    """Wrap the value of `angles_radians` to the range [min_val, max_val]."""
    max_min_diff = max_val - min_val
    return min_val + tf.math.floormod(angle_radians + max_val, max_min_diff)


def _get_3d_rotation_matrix(yaw, roll, pitch):
    """Creates 3x3 rotation matrix from yaw, roll, pitch (angles in radians).

    Note: Yaw -> Z, Roll -> X, Pitch -> Y

    Args:
      yaw: float tensor representing a yaw angle in radians.
      roll: float tensor representing a roll angle in radians.
      pitch: float tensor representing a pitch angle in radians.

    Returns:
      A [3, 3] tensor corresponding to a rotation matrix.

    """

    def _UnitX(angle):
        return tf.reshape(
            [
                1.0,
                0.0,
                0.0,
                0.0,
                tf.cos(angle),
                -tf.sin(angle),
                0.0,
                tf.sin(angle),
                tf.cos(angle),
            ],
            shape=[3, 3],
        )

    def _UnitY(angle):
        return tf.reshape(
            [
                tf.cos(angle),
                0.0,
                tf.sin(angle),
                0.0,
                1.0,
                0.0,
                -tf.sin(angle),
                0.0,
                tf.cos(angle),
            ],
            shape=[3, 3],
        )

    def _UnitZ(angle):
        return tf.reshape(
            [
                tf.cos(angle),
                -tf.sin(angle),
                0.0,
                tf.sin(angle),
                tf.cos(angle),
                0.0,
                0.0,
                0.0,
                1.0,
            ],
            shape=[3, 3],
        )

    return tf.matmul(tf.matmul(_UnitZ(yaw), _UnitX(roll)), _UnitY(pitch))


def _center_xyzWHD_to_corner_xyz(boxes):
    """convert from center format to corner format.
    Args:
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].
    Returns:
      corners: [..., num_boxes, 8, 3] float32 Tensor for 3d corners in
        [x, y, z].
    """
    # relative corners w.r.t to origin point
    # this will return all corners in top-down counter clockwise instead of
    # only left top and bottom right.
    rel_corners = tf.constant(
        [
            [0.5, 0.5, 0.5],  # top
            [-0.5, 0.5, 0.5],  # top
            [-0.5, -0.5, 0.5],  # top
            [0.5, -0.5, 0.5],  # top
            [0.5, 0.5, -0.5],  # bottom
            [-0.5, 0.5, -0.5],  # bottom
            [-0.5, -0.5, -0.5],  # bottom
            [0.5, -0.5, -0.5],  # bottom
        ]
    )

    centers = boxes[..., :3]
    dimensions = boxes[..., 3:6]
    phi_world = boxes[..., 6]
    leading_shapes = _get_shape(boxes)
    cos = tf.cos(phi_world)
    sin = tf.sin(phi_world)
    zero = tf.zeros_like(cos)
    one = tf.ones_like(cos)
    rotations = tf.reshape(
        tf.stack([cos, -sin, zero, sin, cos, zero, zero, zero, one], axis=-1),
        leading_shapes[:-1] + [3, 3],
    )
    # apply the delta to convert from centers to relative corners format
    rel_corners = tf.einsum("...ni,ji->...nji", dimensions, rel_corners)
    # apply rotation matrix on relative corners
    rel_corners = tf.einsum("...nij,...nkj->...nki", rotations, rel_corners)
    # translate back to absolute corners format
    corners = rel_corners + tf.reshape(centers, leading_shapes[:-1] + [1, 3])
    return corners


def _get_shape(tensor):
    tensor = tf.convert_to_tensor(tensor)
    dynamic_shape = tf.shape(tensor)
    if tensor.shape.ndims is None:
        return dynamic_shape
    static_shape = tensor.shape.as_list()
    shapes = [
        static_shape[x] if static_shape[x] is not None else dynamic_shape[x]
        for x in range(tensor.shape.ndims)
    ]
    return shapes


def _is_on_lefthand_side(points, v1, v2):
    """Checks if points lay on a vector direction or to its left.

    Args:
      point: float Tensor of [num_points, 2] of points to check
      v1: float Tensor of [num_points, 2] of starting point of the vector
      v2: float Tensor of [num_points, 2] of ending point of the vector

    Returns:
      a boolean Tensor of [num_points] indicate whether each point is on
      the left of the vector or on the vector direction.
    """
    # Prepare for broadcast: All point operations are on the right,
    # and all v1/v2 operations are on the left. This is faster than left/right
    # under the assumption that we have more points than vertices.
    points_x = points[..., tf.newaxis, :, 0]
    points_y = points[..., tf.newaxis, :, 1]
    v1_x = v1[..., 0, tf.newaxis]
    v2_x = v2[..., 0, tf.newaxis]
    v1_y = v1[..., 1, tf.newaxis]
    v2_y = v2[..., 1, tf.newaxis]
    d1 = (points_y - v1_y) * (v2_x - v1_x)
    d2 = (points_x - v1_x) * (v2_y - v1_y)
    return d1 >= d2


def _box_area(boxes):
    """Compute the area of 2-d boxes.

    Vertices must be ordered counter-clockwise. This function can
    technically handle any kind of convex polygons.

    Args:
      boxes: a float Tensor of [..., 4, 2] of boxes. The last coordinates
        are the four corners of the box and (x, y). The corners must be given in
        counter-clockwise order.
    """
    boxes_roll = tf.roll(boxes, shift=1, axis=-2)
    det = (
        tf.reduce_sum(
            boxes[..., 0] * boxes_roll[..., 1]
            - boxes[..., 1] * boxes_roll[..., 0],
            axis=-1,
            keepdims=True,
        )
        / 2.0
    )
    return tf.abs(det)


def is_within_box2d(points, boxes):
    """Checks if 3d points are within 2d bounding boxes.
    Currently only xy format is supported.
    This function returns true if points are strictly inside the box or on edge.

    Args:
      points: [num_points, 2] float32 Tensor for 2d points in xy format.
      boxes: [num_boxes, 4, 2] float32 Tensor for 2d boxes in xy format,
        counter clockwise.

    Returns:
      boolean Tensor of shape [num_points, num_boxes]
    """
    v1, v2, v3, v4 = (
        boxes[..., 0, :],
        boxes[..., 1, :],
        boxes[..., 2, :],
        boxes[..., 3, :],
    )
    is_inside = tf.math.logical_and(
        tf.math.logical_and(
            _is_on_lefthand_side(points, v1, v2),
            _is_on_lefthand_side(points, v2, v3),
        ),
        tf.math.logical_and(
            _is_on_lefthand_side(points, v3, v4),
            _is_on_lefthand_side(points, v4, v1),
        ),
    )
    valid_area = tf.greater(_box_area(boxes), 0)
    is_inside = tf.math.logical_and(is_inside, valid_area)
    # swap the last two dimensions
    is_inside = tf.einsum("...ij->...ji", tf.cast(is_inside, tf.int32))
    return tf.cast(is_inside, tf.bool)


def is_within_any_box3d(points, boxes, keepdims=False):
    """Checks if 3d points are within any 3d bounding boxes.
    Currently only xyz format is supported.

    Args:
      points: [..., num_points, 3] float32 Tensor for 3d points in xyz format.
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].
      keepdims: boolean. If true, retains reduced dimensions with length 1.

    Returns:
      boolean Tensor of shape [..., num_points] indicating whether
      the point belongs to the box.

    """
    res = is_within_box3d(points, boxes)
    return tf.reduce_any(res, axis=-1, keepdims=keepdims)


def is_within_box3d(points, boxes):
    """Checks if 3d points are within 3d bounding boxes.
    Currently only xyz format is supported.

    Args:
      points: [..., num_points, 3] float32 Tensor for 3d points in xyz format.
      boxes: [..., num_boxes, 7] float32 Tensor for 3d boxes in [x, y, z, dx,
        dy, dz, phi].

    Returns:
      boolean Tensor of shape [..., num_points, num_boxes] indicating whether
      the point belongs to the box.

    """
    # step 1 -- determine if points are within xy range

    # convert from center format to corner format
    boxes_corner = _center_xyzWHD_to_corner_xyz(boxes)
    # project to 2d boxes by only taking x, y on top plane
    boxes_2d = boxes_corner[..., 0:4, 0:2]
    # project to 2d points by only taking x, y
    points_2d = points[..., :2]
    # check whether points are within 2d boxes, [..., num_points, num_boxes]
    is_inside_2d = is_within_box2d(points_2d, boxes_2d)

    # step 2 -- determine if points are within z range

    [_, _, z, _, _, dz, _] = tf.split(boxes, 7, axis=-1)
    z = z[..., 0]
    dz = dz[..., 0]
    bottom = z - dz / 2.0
    # [..., 1, num_boxes]
    bottom = bottom[..., tf.newaxis, :]
    top = z + dz / 2.0
    top = top[..., tf.newaxis, :]
    # [..., num_points, 1]
    points_z = points[..., 2:]
    # [..., num_points, num_boxes]
    is_inside_z = tf.math.logical_and(
        tf.less_equal(points_z, top), tf.greater_equal(points_z, bottom)
    )
    return tf.math.logical_and(is_inside_z, is_inside_2d)


def coordinate_transform(points, pose):
    """
    Translate 'points' to coordinates according to 'pose' vector.
    pose should contain 6 floating point values:
      translate_x, translate_y, translate_z: The translation to apply.
      yaw, roll, pitch: The rotation angles in radians.

    Args:
      points: Float shape [..., 3]: Points to transform to new coordinates.
      pose: Float shape [6]: [translate_x, translate_y, translate_z, yaw, roll,
        pitch]. The pose in the frame that 'points' comes from, and the
        definition of the rotation and translation angles to apply to points.
    Returns:
    'points' transformed to the coordinates defined by 'pose'.
    """
    translate_x = pose[0]
    translate_y = pose[1]
    translate_z = pose[2]

    # Translate the points so the origin is the pose's center.
    translation = tf.reshape([translate_x, translate_y, translate_z], shape=[3])
    translated_points = points + translation

    # Compose the rotations along the three axes.
    #
    # Note: Yaw->Z, Roll->X, Pitch->Y.
    yaw, roll, pitch = pose[3], pose[4], pose[5]
    rotation_matrix = _get_3d_rotation_matrix(yaw, roll, pitch)
    # Finally, rotate the points about the pose's origin according to the
    # rotation matrix.
    rotated_points = tf.einsum(
        "...i,...ij->...j", translated_points, rotation_matrix
    )
    return rotated_points


def spherical_coordinate_transform(points):
    """Converts points from xyz coordinates to spherical coordinates.
    https://en.wikipedia.org/wiki/Spherical_coordinate_system#Coordinate_system_conversions
    for definitions of the transformations.
    Args:
      points: A floating point tensor with shape [..., 3], where the inner 3
        dimensions correspond to xyz coordinates.
    Returns:
      A floating point tensor with the same shape [..., 3], where the inner
      dimensions correspond to (dist, theta, phi), where phi corresponds to
      azimuth/yaw (rotation around z), and theta corresponds to
      pitch/inclination (rotation around y).
    """
    dist = tf.sqrt(tf.reduce_sum(tf.square(points), axis=-1))
    theta = tf.acos(points[..., 2] / tf.maximum(dist, 1e-7))
    # Note: tf.atan2 takes in (y, x).
    phi = tf.atan2(points[..., 1], points[..., 0])
    return tf.stack([dist, theta, phi], axis=-1)


def within_a_frustum(points, center, r_distance, theta_width, phi_width):
    """Check if 3d points are within a 3d frustum.
    https://en.wikipedia.org/wiki/Spherical_coordinate_system for definitions of
    r, theta, and phi. https://en.wikipedia.org/wiki/Viewing_frustum for
    definition of a viewing frustum. Here, we use a conical shaped frustum
    (https://mathworld.wolfram.com/ConicalFrustum.html). Currently, only xyz
    format is supported.

    Args:
      points: [num_points, 3] float32 Tensor for 3d points in xyz format.
      center: [3, ] float32 Tensor for the frustum center in xyz format.
      r_distance: A float scalar sets the starting distance of a frustum.
      theta_width: A float scalar sets the theta width of a frustum.
      phi_width: A float scalar sets the phi width of a frustum.

    Returns:
      boolean Tensor of shape [num_points] indicating whether
      points are within the frustum.

    """
    r, theta, phi = tf.unstack(
        spherical_coordinate_transform(points[:, :3]), axis=-1
    )

    _, center_theta, center_phi = tf.unstack(
        spherical_coordinate_transform(center[tf.newaxis, :]), axis=-1
    )

    theta_half_width = theta_width / 2.0
    phi_half_width = phi_width / 2.0

    # Points within theta and phi width and
    # further than r distance are selected.
    in_theta_width = (theta < (center_theta + theta_half_width)) & (
        theta > (center_theta - theta_half_width)
    )

    in_phi_width = (phi < (center_phi + phi_half_width)) & (
        phi > (center_phi - phi_half_width)
    )

    in_r_distance = r > r_distance
    return in_theta_width & in_phi_width & in_r_distance



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras


@keras_cv_export("keras_cv.models.retinanet.PredictionHead")
class PredictionHead(keras.layers.Layer):
    """The class/box predictions head.

    Arguments:
      output_filters: Number of convolution filters in the final layer.
      bias_initializer: Bias Initializer for the final convolution layer.

    Returns:
      A function representing either the classification
        or the box regression head depending on `output_filters`.
    """

    def __init__(
        self, output_filters, bias_initializer, num_conv_layers=3, **kwargs
    ):
        super().__init__(**kwargs)
        self.output_filters = output_filters
        self.bias_initializer = bias_initializer
        self.num_conv_layers = num_conv_layers

        self.conv_layers = [
            keras.layers.Conv2D(
                256,
                kernel_size=3,
                padding="same",
                kernel_initializer="orthogonal",
                activation="relu",
            )
            for _ in range(num_conv_layers)
        ]
        self.prediction_layer = keras.layers.Conv2D(
            self.output_filters,
            kernel_size=3,
            strides=1,
            padding="same",
            kernel_initializer="orthogonal",
            bias_initializer=self.bias_initializer,
        )

    def call(self, x, training=False):
        for layer in self.conv_layers:
            x = layer(x, training=training)
        x = self.prediction_layer(x, training=training)
        return x

    def compute_output_shape(self, input_shape):
        return tuple(input_shape[:-1]) + (self.output_filters,)

    def get_config(self):
        config = {
            "bias_initializer": keras.initializers.serialize(
                self.bias_initializer
            ),
            "output_filters": self.output_filters,
            "num_conv_layers": self.num_conv_layers,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        config.update(
            {
                "bias_initializer": keras.initializers.deserialize(
                    config["bias_initializer"]
                )
            }
        )
        return super().from_config(config)

    def build(self, input_shape):
        self.conv_layers[0].build(input_shape)

        intermediate_shape = tuple(input_shape[:-1]) + (256,)
        for conv_layer in self.conv_layers[1:]:
            conv_layer.build(intermediate_shape)

        self.prediction_layer.build(intermediate_shape)
        self.built = True



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.backend import random


@keras_cv_export("keras_cv.layers.DropPath")
class DropPath(keras.layers.Layer):
    """
    Implements the DropPath layer. DropPath randomly drops samples during
    training with a probability of `rate`. Note that this layer drops individual
    samples within a batch and not the entire batch. DropPath randomly drops
    some individual samples from a batch, whereas StochasticDepth
    randomly drops the entire batch.

    References:
        - [FractalNet](https://arxiv.org/abs/1605.07648v4).
        - [rwightman/pytorch-image-models](https://github.com/rwightman/pytorch-image-models/blob/7c67d6aca992f039eece0af5f7c29a43d48c00e4/timm/models/layers/drop.py#L135)

    Args:
        rate: float, the probability of the residual branch being dropped.
        seed: (Optional) integer. Used to create a random seed.

    Example:
    `DropPath` can be used in any network as follows:
    ```python

    # (...)
    input = tf.ones((1, 3, 3, 1), dtype=tf.float32)
    residual = keras.layers.Conv2D(1, 1)(input)
    output = keras_cv.layers.DropPath()(input)
    # (...)
    ```
    """  # noqa: E501

    def __init__(self, rate=0.5, seed=None, **kwargs):
        super().__init__(**kwargs)
        self.rate = rate
        self._seed_val = seed
        self.seed = random.SeedGenerator(seed=seed)

    def call(self, x, training=None):
        if self.rate == 0.0 or not training:
            return x
        else:
            batch_size = x.shape[0] or ops.shape(x)[0]
            drop_map_shape = (batch_size,) + (1,) * (len(x.shape) - 1)
            drop_map = ops.cast(
                random.uniform(drop_map_shape, seed=self.seed) > self.rate,
                x.dtype,
            )
            x = x / (1.0 - self.rate)
            x = x * drop_map
            return x

    def get_config(self):
        config = super().get_config()
        config.update({"rate": self.rate, "seed": self._seed_val})
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.layers.object_detection_3d.heatmap_decoder import (
    HeatmapDecoder,
)
from keras_cv.src.models.object_detection_3d.center_pillar_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.models.task import Task
from keras_cv.src.utils.python_utils import classproperty


@keras_cv_export("keras_cv.models.MultiHeadCenterPillar")
class MultiHeadCenterPillar(Task):
    """Multi headed model based on CenterNet heatmap and PointPillar.

    This model builds box classification and regression for each class
    separately. It voxelizes the point cloud feature, applies feature extraction
    on top of voxelized feature, and applies multi-class classification and
    regression heads on the feature map.

    Args:
        backbone: the backbone to apply to voxelized features.
        voxel_net: the voxel_net that takes point cloud feature and convert
            to voxelized features. KerasCV offers a `DynamicVoxelization` layer
            in `keras_cv.layers` which is a reasonable default for most
            detection use cases.
        multiclass_head: A keras.layers.Layer which takes the backbone output
            and returns a dict of heatmap prediction and regression prediction
            per class.
        prediction_decoder: a multi class heatmap prediction decoder that
            returns a dict of decoded boxes, box class, and box confidence score
            per class.
    """

    def __init__(
        self,
        backbone,
        voxel_net,
        multiclass_head,
        prediction_decoder,
        **kwargs,
    ):
        point_xyz = keras.layers.Input((None, 3), name="point_xyz")
        point_feature = keras.layers.Input((None, 4), name="point_feature")
        point_mask = keras.layers.Input(
            (None, 1), name="point_mask", dtype="bool"
        )

        inputs = {
            "point_xyz": point_xyz,
            "point_feature": point_feature,
            "point_mask": point_mask,
        }

        voxel_feature = voxel_net(point_xyz, point_feature, point_mask[..., 0])
        voxel_feature = backbone(voxel_feature)
        predictions = multiclass_head(voxel_feature)

        # A slight hack to get the output names in the model outputs for a
        # functional model.
        for head_name in multiclass_head._head_names:
            predictions[f"box_{head_name}"] = keras.layers.Identity(
                name=f"box_{head_name}"
            )(predictions[head_name])
            predictions[f"heatmap_{head_name}"] = keras.layers.Identity(
                name=f"heatmap_{head_name}"
            )(predictions[head_name])

        super().__init__(inputs=inputs, outputs=predictions, **kwargs)
        self._backbone = backbone
        self._multiclass_head = multiclass_head
        self._prediction_decoder = prediction_decoder
        self._head_names = self._multiclass_head._head_names

    def compile(self, heatmap_loss=None, box_loss=None, **kwargs):
        """Compiles the MultiHeadCenterPillar.

        `compile()` mirrors the standard Keras `compile()` method, but allows
        for specification of heatmap and box-specific losses.

        Args:
            heatmap_loss: a Keras loss to use for heatmap regression.
            box_loss: a Keras loss to use for box regression, or a list of Keras
                losses for box regression, one for each class. If only one loss
                is specified, it will be used for all classes, otherwise exactly
                one loss should be specified per class.
            kwargs: other `keras.Model.compile()` arguments are supported and
                propagated to the `keras.Model` class.
        """
        losses = {}

        if box_loss is not None and not isinstance(box_loss, list):
            box_loss = [
                box_loss for _ in range(self._multiclass_head._num_classes)
            ]
        for i in range(self._multiclass_head._num_classes):
            losses[f"heatmap_class_{i+1}"] = heatmap_loss
            losses[f"box_class_{i+1}"] = box_loss[i]

        super().compile(loss=losses, **kwargs)

    def compute_loss(self, x, y, y_pred, sample_weight=None, **kwargs):
        predictions = y_pred
        targets = y

        y_pred = {}
        y_true = {}
        sample_weight = {}

        for head_name in self._head_names:
            prediction = predictions[head_name]
            heatmap_pred = ops.softmax(prediction[..., :2])[..., 1]
            box_pred = prediction[..., 2:]
            box = targets[head_name]["boxes"]
            heatmap = targets[head_name]["heatmap"]
            index = targets[head_name]["top_k_index"]

            # the prediction returns 2 outputs for background vs object
            y_pred["heatmap_" + head_name] = heatmap_pred
            y_true["heatmap_" + head_name] = heatmap

            # TODO(ianstenbit): loss heatmap threshold should be configurable.
            box_regression_mask = (
                ops.take_along_axis(
                    ops.reshape(heatmap, (heatmap.shape[0], -1)),
                    index[..., 0] * heatmap.shape[1] + index[..., 1],
                    axis=1,
                )
                > 0.95
            )

            box = ops.take_along_axis(
                ops.reshape(box, (ops.shape(box)[0], -1, 7)),
                ops.expand_dims(
                    index[..., 0] * ops.shape(box)[1] + index[..., 1], axis=-1
                ),
                axis=1,
            )

            box_pred = ops.take_along_axis(
                ops.reshape(
                    box_pred,
                    (ops.shape(box_pred)[0], -1, ops.shape(box_pred)[-1]),
                ),
                ops.expand_dims(
                    index[..., 0] * ops.shape(box_pred)[1] + index[..., 1],
                    axis=-1,
                ),
                axis=1,
            )

            box_center_mask = heatmap > 0.99
            num_boxes = ops.maximum(
                ops.sum(ops.cast(box_center_mask, "float32"), axis=[1, 2]), 1
            )

            sample_weight["box_" + head_name] = ops.cast(
                box_regression_mask, "float32"
            ) / ops.broadcast_to(
                ops.expand_dims(num_boxes, axis=-1),
                ops.shape(box_regression_mask),
            )
            sample_weight["heatmap_" + head_name] = ops.ones_like(
                heatmap
            ) / ops.broadcast_to(
                ops.expand_dims(ops.expand_dims(num_boxes, axis=-1), axis=-1),
                heatmap.shape,
            )

            y_pred["box_" + head_name] = box_pred
            y_true["box_" + head_name] = box

        return super().compute_loss(
            x={}, y=y_true, y_pred=y_pred, sample_weight=sample_weight
        )

    def predict_step(self, *args):
        outputs = super().predict_step(*args)
        if isinstance(outputs, tuple):
            return self._prediction_decoder(outputs[0]), outputs[1]
        else:
            return self._prediction_decoder(outputs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(backbone_presets)

    @classproperty
    def backbone_presets(cls):
        """Dictionary of preset names and configurations of compatible
        backbones."""
        return copy.deepcopy(backbone_presets)


class MultiClassDetectionHead(keras.layers.Layer):
    """Multi-class object detection head for CenterPillar.

    This head includes a 1x1 convolution layer for each class which is called
    on the output of the CenterPillar's backbone. The outputs are per-class
    prediction heatmaps which must be decoded into 3D boxes.

    Args:
        num_classes: int, the number of box classes to predict.
        num_head_bin: list of ints, the number of heading bins to use for each
            respective box class.
    """

    def __init__(
        self,
        num_classes,
        num_head_bin,
        name="detection_head",
    ):
        super().__init__(name=name)

        self._heads = {}
        self._head_names = []
        self._num_classes = num_classes
        self._num_head_bin = num_head_bin

        for i in range(num_classes):
            self._head_names.append(f"class_{i + 1}")

            # 1x1 conv for each voxel/pixel.
            self._heads[self._head_names[i]] = keras.layers.Conv2D(
                # 2 for class, 3 for location, 3 for size, 2N for heading
                filters=8 + 2 * num_head_bin[i],
                kernel_size=(1, 1),
                name=f"head_{i + 1}",
            )

    def call(self, feature, training=True):
        del training
        outputs = {}
        for head_name in self._head_names:
            outputs[head_name] = self._heads[head_name](feature)
        return outputs


class MultiClassHeatmapDecoder(keras.layers.Layer):
    """Heatmap decoder for CenterPillar models.

    The heatmap decoder converts a sparse heatmap of box predictions into a
    padded dense set of decoded predicted boxes.

    The input to the heatmap decoder is a spatial heatmap of encoded box
    predictions, and the output is decoded 3D boxes in CENTER_XYZ_DXDYDZ_PHI
    format.

    Args:
        num_classes: int, the number of box classes to predict.
        num_head_bin: list of ints, the number of heading bins for each
            respective class.
        anchor_size: list of length-3 lists of floats, the 3D anchor sizes for
            each respective class.
        max_pool_size: list of ints, the 2D pooling size for the heatmap, to be
            used before box decoding.
        max_num_box: list of ints, the maximum number of boxes to return for
            each class. The top K boxes will be returned, and if fewer than K
            boxes are predicted, the outputs will be padded to contain K boxes.
        heatmap_threshold: list of floats, the heatmap confidence threshold to
            be used for each respective class to determine whether or not a box
            prediction is strong enough to decode and return.
        voxel_size: list of floats, the size of the voxels that were used to
            voxelize inputs to the CenterPillar model for each respective class.
        spatial_size: list of floats, the global 3D size of the heatmap for each
            respective class. `spatial_size[i] / voxel_size[i]` equals the
            size of the `i`th rank of the input heatmap.
    """

    def __init__(
        self,
        num_classes,
        num_head_bin,
        anchor_size,
        max_pool_size,
        max_num_box,
        heatmap_threshold,
        voxel_size,
        spatial_size,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_classes = num_classes
        self.class_ids = list(range(1, num_classes + 1))
        self.num_head_bin = num_head_bin
        self.anchor_size = anchor_size
        self.max_pool_size = max_pool_size
        self.max_num_box = max_num_box
        self.heatmap_threshold = heatmap_threshold
        self.voxel_size = voxel_size
        self.spatial_size = spatial_size
        self.decoders = {}
        for i, class_id in enumerate(self.class_ids):
            self.decoders[f"class_{class_id}"] = HeatmapDecoder(
                class_id=class_id,
                num_head_bin=self.num_head_bin[i],
                anchor_size=self.anchor_size[i],
                max_pool_size=self.max_pool_size[i],
                max_num_box=self.max_num_box[i],
                heatmap_threshold=self.heatmap_threshold[i],
                voxel_size=self.voxel_size,
                spatial_size=self.spatial_size,
            )

    def call(self, predictions):
        box_predictions = []
        class_predictions = []
        box_confidence = []
        for class_id in self.class_ids:
            class_tag = f"class_{class_id}"
            boxes, classes, confidence = self.decoders[class_tag](
                predictions[class_tag]
            )
            box_predictions.append(boxes)
            class_predictions.append(classes)
            box_confidence.append(confidence)

        return {
            "3d_boxes": {
                "boxes": ops.concatenate(box_predictions, axis=1),
                "classes": ops.concatenate(class_predictions, axis=1),
                "confidence": ops.concatenate(box_confidence, axis=1),
            }
        }



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import functools

from keras_cv.src import backend
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.backend import tf_ops
from keras_cv.src.backend.config import keras_3

_ORIGINAL_OPS = copy.copy(backend.ops.__dict__)
_ORIGINAL_SUPPORTS_RAGGED = backend.supports_ragged

# A counter for potentially nested TF data scopes
_IN_TF_DATA_SCOPE = 0


def tf_data(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        if keras_3() and keras.src.utils.backend_utils.in_tf_graph():
            with TFDataScope():
                return function(*args, **kwargs)
        else:
            return function(*args, **kwargs)

    return wrapper


class TFDataScope:
    def __enter__(self):
        global _IN_TF_DATA_SCOPE
        if _IN_TF_DATA_SCOPE == 0:
            for k, v in ops.__dict__.items():
                if k in tf_ops.__dict__:
                    setattr(ops, k, getattr(tf_ops, k))
            backend.supports_ragged = lambda: True
        _IN_TF_DATA_SCOPE += 1

    def __exit__(self, exc_type, exc_value, exc_tb):
        global _IN_TF_DATA_SCOPE
        _IN_TF_DATA_SCOPE -= 1
        if _IN_TF_DATA_SCOPE == 0:
            for k, v in ops.__dict__.items():
                setattr(ops, k, _ORIGINAL_OPS[k])
            backend.supports_ragged = _ORIGINAL_SUPPORTS_RAGGED
            _IN_TF_DATA_SCOPE = False



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MiT backbone model.

References:
  - [SegFormer: Simple and Efficient Design for Semantic Segmentation with Transformers](https://arxiv.org/abs/2105.15203) (CVPR 2021)
  - [Based on the TensorFlow implementation from DeepVision](https://github.com/DavidLandup0/deepvision/blob/main/deepvision/models/classification/mix_transformer/mit_tf.py)
  - [Based on the NVlabs' official PyTorch implementation](https://github.com/NVlabs/SegFormer/blob/master/mmseg/models/backbones/mix_transformer.py)
  - [Inspired by @sithu31296's reimplementation](https://github.com/sithu31296/semantic-segmentation/blob/main/semseg/models/backbones/mit.py)
"""  # noqa: E501

import copy

import numpy as np

from keras_cv.src import layers as cv_layers
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models import utils
from keras_cv.src.models.backbones.backbone import Backbone
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_backbone_presets import (  # noqa: E501
    backbone_presets_with_weights,
)
from keras_cv.src.utils.python_utils import classproperty


@keras_cv_export("keras_cv.models.MiTBackbone")
class MiTBackbone(Backbone):
    def __init__(
        self,
        include_rescaling,
        depths,
        input_shape=(224, 224, 3),
        input_tensor=None,
        embedding_dims=None,
        **kwargs,
    ):
        """A Keras model implementing the MixTransformer architecture to be
        used as a backbone for the SegFormer architecture.

        References:
            - [SegFormer: Simple and Efficient Design for Semantic Segmentation with Transformers](https://arxiv.org/abs/2105.15203) # noqa: E501
            - [Based on the TensorFlow implementation from DeepVision](https://github.com/DavidLandup0/deepvision/tree/main/deepvision/models/classification/mix_transformer) # noqa: E501

        Args:
            include_rescaling: bool, whether to rescale the inputs. If set
                to `True`, inputs will be passed through a `Rescaling(1/255.0)`
                layer.
            depths: the number of transformer encoders to be used per stage in the
                network
            embedding_dims: the embedding dims per hierarchical stage, used as
                the levels of the feature pyramid
            input_shape: optional shape tuple, defaults to (None, None, 3).
            input_tensor: optional Keras tensor (i.e. output of `keras.layers.Input()`)
                to use as image input for the model.

        Example:

        Using the class with a `backbone`:

        ```python
        import tensorflow as tf
        import keras_cv

        images = np.ones(shape=(1, 96, 96, 3))
        labels = np.zeros(shape=(1, 96, 96, 1))
        backbone = keras_cv.models.MiTBackbone.from_preset("mit_b0_imagenet")

        # Evaluate model
        model(images)

        # Train model
        model.compile(
            optimizer="adam",
            loss=keras.losses.BinaryCrossentropy(from_logits=False),
            metrics=["accuracy"],
        )
        model.fit(images, labels, epochs=3)
        ```
        """
        drop_path_rate = 0.1
        dpr = [x for x in np.linspace(0.0, drop_path_rate, sum(depths))]
        blockwise_num_heads = [1, 2, 5, 8]
        blockwise_sr_ratios = [8, 4, 2, 1]
        num_stages = 4

        cur = 0
        patch_embedding_layers = []
        transformer_blocks = []
        layer_norms = []

        for i in range(num_stages):
            patch_embed_layer = cv_layers.OverlappingPatchingAndEmbedding(
                project_dim=embedding_dims[0] if i == 0 else embedding_dims[i],
                patch_size=7 if i == 0 else 3,
                stride=4 if i == 0 else 2,
                name=f"patch_and_embed_{i}",
            )
            patch_embedding_layers.append(patch_embed_layer)

            transformer_block = [
                cv_layers.HierarchicalTransformerEncoder(
                    project_dim=embedding_dims[i],
                    num_heads=blockwise_num_heads[i],
                    sr_ratio=blockwise_sr_ratios[i],
                    drop_prob=dpr[cur + k],
                    name=f"hierarchical_encoder_{i}_{k}",
                )
                for k in range(depths[i])
            ]
            transformer_blocks.append(transformer_block)
            cur += depths[i]
            layer_norms.append(keras.layers.LayerNormalization())

        inputs = utils.parse_model_inputs(input_shape, input_tensor)
        x = inputs

        if include_rescaling:
            x = keras.layers.Rescaling(scale=1 / 255)(x)

        pyramid_level_inputs = []
        for i in range(num_stages):
            # Compute new height/width after the `proj`
            # call in `OverlappingPatchingAndEmbedding`
            stride = 4 if i == 0 else 2
            new_height, new_width = (
                int(ops.shape(x)[1] / stride),
                int(ops.shape(x)[2] / stride),
            )

            x = patch_embedding_layers[i](x)
            for blk in transformer_blocks[i]:
                x = blk(x)
            x = layer_norms[i](x)
            x = keras.layers.Reshape(
                (new_height, new_width, -1), name=f"output_level_{i}"
            )(x)
            pyramid_level_inputs.append(utils.get_tensor_input_name(x))

        super().__init__(inputs=inputs, outputs=x, **kwargs)

        self.depths = depths
        self.embedding_dims = embedding_dims
        self.include_rescaling = include_rescaling
        self.input_tensor = input_tensor
        self.pyramid_level_inputs = {
            f"P{i + 1}": name for i, name in enumerate(pyramid_level_inputs)
        }

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "depths": self.depths,
                "embedding_dims": self.embedding_dims,
                "include_rescaling": self.include_rescaling,
                "input_shape": self.input_shape[1:],
                "input_tensor": self.input_tensor,
            }
        )
        return config

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(backbone_presets)

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy(backbone_presets_with_weights)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.models.stable_diffusion.clip_tokenizer import SimpleTokenizer
from keras_cv.src.models.stable_diffusion.decoder import Decoder
from keras_cv.src.models.stable_diffusion.diffusion_model import DiffusionModel
from keras_cv.src.models.stable_diffusion.image_encoder import ImageEncoder
from keras_cv.src.models.stable_diffusion.noise_scheduler import NoiseScheduler
from keras_cv.src.models.stable_diffusion.text_encoder import TextEncoder
from keras_cv.src.models.stable_diffusion.text_encoder import TextEncoderV2



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing


@keras_cv_export("keras_cv.layers.Solarization")
class Solarization(VectorizedBaseImageAugmentationLayer):
    """Applies (max_value - pixel + min_value) for each pixel in the image.

    When created without `threshold` parameter, the layer performs solarization
    to all values. When created with specified `threshold` the layer only
    augments pixels that are above the `threshold` value

    Reference:
    - [AutoAugment: Learning Augmentation Policies from Data](
        https://arxiv.org/abs/1805.09501
    )
    - [RandAugment](https://arxiv.org/pdf/1909.13719.pdf)

    Args:
        value_range: a tuple or a list of two elements. The first value
            represents the lower bound for values in passed images, the second
            represents the upper bound. Images passed to the layer should have
            values within `value_range`.
        addition_factor: (Optional)  A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is
            sampled from the provided range. If a float is passed, the range is
            interpreted as `(0, addition_factor)`. If specified, this value is
            added to each pixel before solarization and thresholding. The
            addition value should be scaled according to the value range
            (0, 255), defaults to 0.0.
        threshold_factor: (Optional)  A tuple of two floats, a single float or
            a `keras_cv.FactorSampler`. For each augmented image a value is
            sampled from the provided range. If a float is passed, the range is
            interpreted as `(0, threshold_factor)`. If specified, only pixel
            values above this threshold will be solarized.
        seed: Integer. Used to create a random seed.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    print(images[0, 0, 0])
    # [59 62 63]
    # Note that images are Tensor with values in the range [0, 255]
    solarization = Solarization(value_range=(0, 255))
    images = solarization(images)
    print(images[0, 0, 0])
    # [196, 193, 192]
    ```

    Call arguments:
        images: Tensor of type int or float, with pixels in
            range [0, 255] and shape [batch, height, width, channels]
            or [height, width, channels].
    """

    def __init__(
        self,
        value_range,
        addition_factor=0.0,
        threshold_factor=0.0,
        seed=None,
        **kwargs
    ):
        super().__init__(seed=seed, **kwargs)
        self.seed = seed
        self.addition_factor = preprocessing.parse_factor(
            addition_factor,
            max_value=255,
            seed=seed,
            param_name="addition_factor",
        )
        self.threshold_factor = preprocessing.parse_factor(
            threshold_factor,
            max_value=255,
            seed=seed,
            param_name="threshold_factor",
        )
        self.value_range = value_range

    def get_random_transformation_batch(self, batch_size, **kwargs):
        return {
            "additions": self.addition_factor(
                shape=(batch_size, 1, 1, 1), dtype=self.compute_dtype
            ),
            "thresholds": self.threshold_factor(
                shape=(batch_size, 1, 1, 1), dtype=self.compute_dtype
            ),
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(image, transformation)

    def augment_images(self, images, transformations, **kwargs):
        thresholds = transformations["thresholds"]
        additions = transformations["additions"]
        images = preprocessing.transform_value_range(
            images,
            original_range=self.value_range,
            target_range=(0, 255),
            dtype=self.compute_dtype,
        )
        results = images + additions
        results = tf.clip_by_value(results, 0, 255)
        results = tf.where(results < thresholds, results, 255 - results)
        results = preprocessing.transform_value_range(
            results,
            original_range=(0, 255),
            target_range=self.value_range,
            dtype=self.compute_dtype,
        )
        return results

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def get_config(self):
        config = {
            "threshold_factor": self.threshold_factor,
            "addition_factor": self.addition_factor,
            "value_range": self.value_range,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        if isinstance(config["threshold_factor"], dict):
            config["threshold_factor"] = keras.utils.deserialize_keras_object(
                config["threshold_factor"]
            )
        if isinstance(config["addition_factor"], dict):
            config["addition_factor"] = keras.utils.deserialize_keras_object(
                config["addition_factor"]
            )
        return cls(**config)
